#!/usr/bin/python
# lookup.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (lookup.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: lookup.

.. role:: red
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

import sys
import time
import random

try:
    from twisted.internet import reactor  # @UnresolvedImport
except:
    sys.exit('Error initializing twisted.internet.reactor in lookup.py')

from twisted.internet.defer import DeferredList, Deferred

#------------------------------------------------------------------------------

from bitdust.lib import strng

from bitdust.logs import lg

from bitdust.dht import dht_service

from bitdust.contacts import identitycache

from bitdust.userid import id_url

#------------------------------------------------------------------------------

_KnownIDURLsDict = {}
_DiscoveredIDURLsList = {}
_LookupTasks = []
_LatestLookupID = 0
_CurrentLookupTask = None
_LookupMethod = None  # method to get a list of random nodes
_ObserveMethod = None  # method to get IDURL from given node
_ProcessMethod = None  # method to do some stuff with discovered IDURL

#------------------------------------------------------------------------------


def init(lookup_method=None, observe_method=None, process_method=None):
    global _LookupMethod
    global _ObserveMethod
    global _ProcessMethod
    _LookupMethod = lookup_method
    _ObserveMethod = observe_method
    _ProcessMethod = process_method
    if _Debug:
        lg.out(_DebugLevel, 'lookup.init')


def shutdown():
    global _LookupMethod
    global _ObserveMethod
    global _ProcessMethod
    _LookupMethod = None
    _ObserveMethod = None
    _ProcessMethod = None
    if _Debug:
        lg.out(_DebugLevel, 'lookup.shutdown')


#------------------------------------------------------------------------------


def known_idurls():
    global _KnownIDURLsDict
    return _KnownIDURLsDict


def discovered_idurls(layer_id=0):
    global _DiscoveredIDURLsList
    if layer_id not in _DiscoveredIDURLsList:
        _DiscoveredIDURLsList[layer_id] = []
    return _DiscoveredIDURLsList[layer_id]


#------------------------------------------------------------------------------


def consume_discovered_idurls(count=1, layer_id=0):
    if not discovered_idurls(layer_id=layer_id):
        if _Debug:
            lg.out(_DebugLevel, 'lookup.consume_discovered_idurls returns empty list')
        return []
    results = []
    while len(results) < count and discovered_idurls(layer_id=layer_id):
        # random_pos = random.randint(0, len(discovered_idurls(layer_id=layer_id)) - 1)
        # results.append(id_url.to_bin(discovered_idurls(layer_id=layer_id).pop(random_pos)))
        results.append(id_url.to_bin(discovered_idurls(layer_id=layer_id).pop(0)))
    if _Debug:
        lg.out(_DebugLevel, 'lookup.consume_discovered_idurls : %s' % results)
    return results


def extract_discovered_idurls(count=1, layer_id=0):
    if not discovered_idurls(layer_id=layer_id):
        if _Debug:
            lg.out(_DebugLevel, 'lookup.extract_discovered_idurls returns empty list')
        return []
    discovered = list(discovered_idurls(layer_id=layer_id))
    # random.shuffle(discovered)
    results = id_url.to_bin_list(discovered[:count])
    if _Debug:
        lg.out(_DebugLevel, 'lookup.extract_discovered_idurls : %s' % results)
    return results


#------------------------------------------------------------------------------


def random_proxy_router(**kwargs):
    from bitdust.dht import dht_records
    kwargs['layer_id'] = dht_records.LAYER_PROXY_ROUTERS
    return start(**kwargs)


def random_supplier(**kwargs):
    from bitdust.dht import dht_records
    kwargs['layer_id'] = dht_records.LAYER_SUPPLIERS
    return start(**kwargs)


def random_message_broker(**kwargs):
    from bitdust.dht import dht_records
    kwargs['layer_id'] = dht_records.LAYER_MESSAGE_BROKERS
    return start(**kwargs)


def random_merchant(**kwargs):
    from bitdust.dht import dht_records
    kwargs['layer_id'] = dht_records.LAYER_MERCHANTS
    return start(**kwargs)


def random_customer(**kwargs):
    from bitdust.dht import dht_records
    kwargs['layer_id'] = dht_records.LAYER_CUSTOMERS
    return start(**kwargs)


#------------------------------------------------------------------------------


def start(count=1, consume=True, lookup_method=None, observe_method=None, process_method=None, force_discovery=False, ignore_idurls=[], layer_id=0):
    """
    NOTE: no parallel threads, DHT lookup can be started only one at time.
    """
    global _LookupTasks
    t = DiscoveryTask(
        count=count,
        consume=consume,
        lookup_method=lookup_method,
        observe_method=observe_method,
        process_method=process_method,
        ignore_idurls=ignore_idurls,
        layer_id=layer_id,
    )
    if not force_discovery and len(discovered_idurls(layer_id=layer_id)) > count:
        if _Debug:
            lg.out(_DebugLevel - 4, 'lookup.start  knows %d discovered nodes, SKIP and return %d nodes' % (len(discovered_idurls(layer_id=layer_id)), count))
        if consume:
            idurls = consume_discovered_idurls(count, layer_id=layer_id)
        else:
            idurls = extract_discovered_idurls(count, layer_id=layer_id)
        reactor.callLater(0, t.result_defer.callback, idurls)  # @UndefinedVariable
        return t
    # if force_discovery:
    #     discovered_idurls(layer_id=layer_id).clear()
    _LookupTasks.append(t)
    reactor.callLater(0, work)  # @UndefinedVariable
    if _Debug:
        lg.out(_DebugLevel - 4, 'lookup.start  new DiscoveryTask created for %d nodes at layer %d' % (count, layer_id))
    return t


#------------------------------------------------------------------------------


def on_lookup_task_success(result):
    global _CurrentLookupTask
    if _Debug:
        lg.out(_DebugLevel - 4, 'lookup.on_lookup_task_success %s' % result)
    _CurrentLookupTask = None
    reactor.callLater(0, work)  # @UndefinedVariable
    return result


def on_lookup_task_failed(err):
    global _CurrentLookupTask
    if _Debug:
        lg.out(_DebugLevel - 4, 'lookup.on_lookup_task_failed: %s' % err)
    _CurrentLookupTask = None
    reactor.callLater(0, work)  # @UndefinedVariable
    return err


def work():
    global _CurrentLookupTask
    global _LookupTasks
    if _CurrentLookupTask:
        if _Debug:
            lg.out(_DebugLevel - 4, 'lookup.work SKIP, %s is in progress' % _CurrentLookupTask)
        return
    if not _LookupTasks:
        if _Debug:
            lg.out(_DebugLevel - 4, 'lookup.work SKIP no lookup tasks in the queue')
        return
    if _Debug:
        lg.out(_DebugLevel - 4, 'lookup.work starting next task in the queue')
    _CurrentLookupTask = _LookupTasks.pop(0)
    if _CurrentLookupTask.stopped or not _CurrentLookupTask.result_defer or not _CurrentLookupTask.lookup_method:
        lg.warn('task %s is closed' % _CurrentLookupTask)
        _CurrentLookupTask = None
        return
    _CurrentLookupTask.start()
    if _CurrentLookupTask.result_defer:
        _CurrentLookupTask.result_defer.addCallback(on_lookup_task_success)
        _CurrentLookupTask.result_defer.addErrback(on_lookup_task_failed)
    else:
        lg.warn('task %s was closed immediately' % _CurrentLookupTask)
        _CurrentLookupTask = None


#------------------------------------------------------------------------------


def lookup_in_dht(layer_id=0):
    """
    Pick random node from Distributed Hash Table.
    Generates
    """
    if _Debug:
        lg.out(_DebugLevel, 'lookup.lookup_in_dht layer_id=%d' % (layer_id, ))
    d = dht_service.find_node(dht_service.random_key(), layer_id=layer_id)
    d.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='lookup_in_dht')
    return d


def on_idurl_response(response, result):
    if _Debug:
        lg.out(_DebugLevel, 'lookup.on_idurl_response : %r' % response)
    responded_idurl = response.get('idurl')
    if not responded_idurl:
        result.errback(Exception('idurl observe failed'))
        return response
    try:
        idurl = id_url.to_bin(responded_idurl)
    except:
        lg.exc()
        result.errback(Exception('idurl observe failed'))
        return response
    result.callback(idurl)
    return response


def observe_dht_node(node, layer_id=0):
    if _Debug:
        lg.out(_DebugLevel, 'lookup.observe_dht_node   %s  layer_id=%d' % (node, layer_id))
    result = Deferred()
    d = node.request('idurl', layerID=layer_id)
    d.addCallback(on_idurl_response, result)
    d.addErrback(result.errback)
    return result


def on_identity_cached(src, idurl, result):
    if not src:
        if _Debug:
            lg.out(_DebugLevel, 'lookup.on_identity_cached FAILED with empty identity source for %r' % idurl)
        result.errback(Exception(idurl))
        return None
    if _Debug:
        lg.out(_DebugLevel, 'lookup.on_identity_cached %r with %d bytes' % (idurl, len(src)))
    result.callback(id_url.field(idurl))
    return src


def process_idurl(idurl, node):
    if _Debug:
        lg.out(_DebugLevel, 'lookup.process_idurl %r from %r' % (idurl, node))
    result = Deferred()
    if not idurl:
        result.errback(Exception(idurl))
        return result
    d = identitycache.immediatelyCaching(idurl)
    d.addCallback(on_identity_cached, idurl, result)
    d.addErrback(result.errback)
    return result


#------------------------------------------------------------------------------


class DiscoveryTask(object):
    def __init__(
        self,
        count,
        consume=True,
        lookup_method=None,
        observe_method=None,
        process_method=None,
        ignore_idurls=[],
        layer_id=0,
    ):
        global _LookupMethod
        global _ObserveMethod
        global _ProcessMethod
        global _LatestLookupID
        _LatestLookupID += 1
        self.id = _LatestLookupID
        self.lookup_method = lookup_method or _LookupMethod
        self.observe_method = observe_method or _ObserveMethod
        self.process_method = process_method or _ProcessMethod
        self.ignore_idurls = ignore_idurls
        self.started = time.time()
        self.layer_id = layer_id
        self.count = count
        self.consume = consume
        self.observed_count = 0
        self.cached_count = 0
        self.succeed = 0
        self.failed = 0
        self.lookup_now = False
        self.stopped = False
        self.observe_finished = False
        self.lookup_task = None
        self.result_defer = Deferred(canceller=lambda d: self._close())
        self.result_defer.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='DiscoveryTask')
        if _Debug:
            lg.out(_DebugLevel, 'lookup.DiscoveryTask[%r].__init__   layer_id=%d' % (self.id, self.layer_id))

    def __del__(self):
        if _Debug:
            lg.out(_DebugLevel, 'lookup.DiscoveryTask[%r].__del__   layer_id=%d' % (self.id, self.layer_id))

    def start(self):
        if self.stopped:
            lg.warn('DiscoveryTask[%r] : task already stopped' % self.id)
            return None
        if self.lookup_task and not self.lookup_task.called:
            lg.warn('DiscoveryTask[%r] : lookup_nodes() method already called' % self.id)
            return self.lookup_task
        if _Debug:
            lg.out(_DebugLevel, 'lookup.DiscoveryTask[%r].start  layer_id=%d' % (self.id, self.layer_id))
        return self._lookup_nodes()

    def stop(self):
        if _Debug:
            lg.out(_DebugLevel, 'lookup.DiscoveryTask[%r].stop' % self.id)
        self.stopped = True
        self._close()

    def _close(self):
        self.stopped = True
        if self.lookup_task and not self.lookup_task.called:
            self.lookup_task.cancel()
            self.lookup_task = None
        self.lookup_method = None
        self.observe_method = None
        self.process_method = None
        self.result_defer = None
        if _Debug:
            lg.out(_DebugLevel, 'lookup.DiscoveryTask[%r].close finished in %f seconds' % (self.id, round(time.time() - self.started, 3)))

    def _lookup_nodes(self):
        if self.lookup_task and not self.lookup_task.called:
            if _Debug:
                lg.out(_DebugLevel, 'lookup.DiscoveryTask[%r]._lookup_nodes    SKIP, already started' % self.id)
            return self.lookup_task
        if _Debug:
            lg.out(_DebugLevel, 'lookup.DiscoveryTask[%r]._lookup_nodes layer_id=%d' % (self.id, self.layer_id))
        self.lookup_now = True
        self.lookup_task = self.lookup_method(layer_id=self.layer_id)
        self.lookup_task.addCallback(self._on_nodes_discovered)
        self.lookup_task.addErrback(self._on_lookup_failed)
        return self.lookup_task

    def _observe_nodes(self, nodes):
        if self.stopped:
            lg.warn('DiscoveryTask[%r] : discovery process already stopped' % self.id)
            return
        if _Debug:
            lg.out(_DebugLevel, 'lookup.DiscoveryTask[%r]._observe_nodes  started for %d items  layer_id=%d' % (self.id, len(nodes), self.layer_id))
        observe_list = []
        for node in nodes:
            d = self.observe_method(node, layer_id=self.layer_id)
            d.addCallback(self._on_node_observed, node)
            d.addErrback(self._on_node_observe_failed, node)
            observe_list.append(d)
        self.observed_count = len(nodes)
        dl = DeferredList(observe_list, consumeErrors=False)
        dl.addCallback(self._on_all_nodes_observed)
        dl.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='DiscoveryTask._observe_nodes')

    def _report_result(self, results=None):
        if _Debug:
            lg.out(_DebugLevel, 'lookup.DiscoveryTask[%r]_report_result in %f seconds   %s,   result_defer=%s' % (self.id, round(time.time() - self.started, 3), str(results), self.result_defer))
        if results is None:
            if self.consume:
                results = consume_discovered_idurls(self.count, layer_id=self.layer_id)
                if _Debug:
                    lg.out(_DebugLevel, 'lookup.DiscoveryTask[%r]    %d results consumed, %d were requested' % (self.id, len(results), self.count))
            else:
                results = extract_discovered_idurls(self.count, layer_id=self.layer_id)
                if _Debug:
                    lg.out(_DebugLevel, 'lookup.DiscoveryTask[%r]    %d results extracted, %d were requested' % (self.id, len(results), self.count))
        if self.result_defer and not self.result_defer.called:
            self.result_defer.callback(results)
        self.result_defer = None

    def _report_fails(self, err):
        lg.err('DHT lookup %r failed: %r' % (self.id, err))
        if self.result_defer:
            self.result_defer.errback(err)
        self.result_defer = None

    def _on_node_succeed(self, node, idurl):
        self.succeed += 1
        if _Debug:
            lg.out(_DebugLevel, 'lookup.lookup.DiscoveryTask[%r]._on_node_succeed %r : %r' % (self.id, node, idurl))
        reactor.callLater(0, self._on_node_processed, node, idurl)  # @UndefinedVariable
        return node

    def _on_node_process_failed(self, err, node):
        self.failed += 1
        if _Debug:
            lg.warn('DiscoveryTask[%r] : node %r processing failed with  %r' % (self.id, node, err))
        reactor.callLater(0, self._on_node_processed, node, None)  # @UndefinedVariable
        return None

    def _on_node_observe_failed(self, err, node):
        try:
            self.failed += 1
            if _Debug:
                err = strng.to_text(err, errors='ignore')
                if err.count('idurl observe failed'):
                    err = 'idurl observe failed'
                lg.args(_DebugLevel, node=node, err=err, task_id=self.id)
        except:
            lg.exc()
        return None

    def _on_node_observed(self, idurl, node):
        if self.stopped:
            lg.warn('DiscoveryTask[%r] : node observed, but discovery process already stopped' % self.id)
            return None
        idurl = id_url.to_bin(idurl)
        if _Debug:
            lg.out(_DebugLevel + 4, 'lookup.DiscoveryTask[%r]._on_node_observed %r : %r' % (self.id, node, idurl))
        cached_time = known_idurls().get(idurl)
        if cached_time and time.time() - cached_time < 30.0:
            if _Debug:
                lg.out(_DebugLevel + 4, 'lookup.DiscoveryTask[%r]._on_node_observed   SKIP processing node %r because already observed recently' % (self.id, idurl))
            self._on_identity_cached(idurl, node)
            return idurl
        d = self.process_method(idurl, node)
        d.addCallback(self._on_identity_cached, node)
        d.addErrback(self._on_node_process_failed, node)
        return idurl

    def _on_node_processed(self, node, idurl):
        if self.stopped:
            if _Debug:
                lg.out(_DebugLevel, 'lookup.DiscoveryTask[%r]._on_node_processed   node %s processed but task already finished' % (self.id, idurl))
            return None
        if _Debug:
            lg.out(_DebugLevel, 'lookup.DiscoveryTask[%r]._on_node_processed  %r  discovered_idurls=%d count=%d  idurl=%s' % (self.id, node, len(discovered_idurls(layer_id=self.layer_id)), self.count, idurl))
        if self.succeed + self.failed >= self.count:
            if _Debug:
                lg.out(_DebugLevel, 'lookup.DiscoveryTask[%r]._on_node_processed   enough node processed : succeed=%d  failed=%d' % (self.id, self.succeed, self.failed))
            self._report_result()
            # self._close()
            return node
        if self.succeed + self.failed >= self.observed_count:
            if _Debug:
                lg.out(_DebugLevel, 'lookup.DiscoveryTask[%r]._on_node_processed   all observed nodes are processed : succeed=%d  failed=%d observed_count=%d' % (self.id, self.succeed, self.failed, self.observed_count))
            self._report_result()
            self._close()
            return node
        if self.observe_finished:
            lg.warn('observe process already finished')
            self._report_result()
            self._close()
        return node

    def _on_all_nodes_observed(self, observe_results):
        if self.stopped:
            lg.warn('DiscoveryTask[%r] : observe finished, but discovery process already stopped' % self.id)
            return
        if _Debug:
            lg.out(_DebugLevel, 'lookup.DiscoveryTask[%r]._on_all_nodes_observed results: %r, discovered nodes: %d' % (self.id, observe_results, len(discovered_idurls(layer_id=self.layer_id))))
        self.observe_finished = True
        found_any_nodes = False
        for one_result in observe_results:
            if one_result[0] and one_result[1]:
                found_any_nodes = True
        if not found_any_nodes:
            lg.warn('DiscoveryTask[%r] : did not observed any nodes' % self.id)
            self._report_result()
            self._close()

    def _on_identity_cached(self, idurl, node):
        if self.stopped:
            return None
        if not idurl:
            self._on_node_process_failed(None, node)
            return None
        if id_url.is_in(idurl, self.ignore_idurls):
            if _Debug:
                lg.dbg(_DebugLevel, 'lookup.DiscoveryTask[%r]._on_identity_cached IGNORE %r' % (self.id, idurl))
            self._on_node_process_failed(None, node)
            return None
        self.cached_count += 1
        idurl = id_url.to_bin(idurl)
        if idurl not in discovered_idurls(layer_id=self.layer_id):
            discovered_idurls(layer_id=self.layer_id).append(idurl)
        known_idurls()[idurl] = time.time()
        self._on_node_succeed(node, idurl)
        if _Debug:
            lg.out(_DebugLevel, 'lookup.DiscoveryTask[%r]._on_identity_cached : %s' % (self.id, idurl))
        return idurl

    def _on_nodes_discovered(self, nodes):
        if _Debug:
            lg.out(_DebugLevel, 'lookup.DiscoveryTask[%r]._on_nodes_discovered : %s, stopped=%s' % (self.id, str(nodes), self.stopped))
        self.lookup_now = False
        if self.stopped:
            lg.warn('DiscoveryTask[%r] : nodes are discovered, but the task was already stopped' % self.id)
            self._close()
            return None
        if not nodes:
            self._report_result()
            self._close()
            return None
        random.shuffle(nodes)
        self._observe_nodes(nodes)
        return None

    def _on_lookup_failed(self, err):
        if _Debug:
            lg.out(_DebugLevel, 'lookup.DiscoveryTask[%r]._on_lookup_failed %s' % (self.id, err))
        self.lookup_now = False
        if self.stopped:
            lg.warn('DiscoveryTask[%r] : discovery process already stopped' % self.id)
            self._close()
            return
        self._report_fails(err)
        self._close()
        return err


#------------------------------------------------------------------------------



#!/usr/bin/env python
# constants.py
#
# Copyright (C) 2007-2008 Francois Aucamp, Meraka Institute, CSIR
# See AUTHORS for all authors and contact information.
#
# License: GNU Lesser General Public License, version 3 or later; see COPYING
#          included in this archive for details.
#
# This library is free software, distributed under the terms of
# the GNU Lesser General Public License Version 3, or any later version.
# See the COPYING file included in this archive
#
# The docstrings in this module contain epytext markup; API documentation
# may be created by processing this file with epydoc: http://epydoc.sf.net
"""
This module defines the charaterizing constants of the Kademlia network.

C{checkRefreshInterval} and C{udpDatagramMaxSize} are implementation-
specific constants, and do not affect general Kademlia operation.
"""

######### KADEMLIA CONSTANTS ###########

#: Small number Representing the degree of parallelism in network calls
alpha = 4

#: Maximum number of contacts stored in a bucket; this should be an even number
k = 4

#: Timeout for network operations (in seconds)
rpcTimeout = 10

# Delay between iterations of iterative node lookups (for loose parallelism)  (in seconds)
iterativeLookupDelay = rpcTimeout/2

#: If a k-bucket has not been used for this amount of time, refresh it (in seconds)
refreshTimeout = 600  # 10 min
#: The interval at which nodes replicate (republish/refresh) data they are holding
replicateInterval = refreshTimeout
# The time it takes for data to expire in the network; the original publisher of the data
# will also republish the data at this time if it is still valid
dataExpireTimeout = 86400  # 24 hours
# Default value for all records to be expired
dataExpireSecondsDefaut = 60*60*12  # 12 hours

######## IMPLEMENTATION-SPECIFIC CONSTANTS ###########

#: The interval in which the node should check its whether any buckets need refreshing,
#: or whether any data needs to be republished (in seconds)
checkRefreshInterval = refreshTimeout/5

#: Max size of a single UDP datagram, in bytes. If a message is larger than this, it will
#: be spread accross several UDP packets.
udpDatagramMaxSize = 8192  # 8 KB



#!/usr/bin/env python
# handshaker.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (handshaker.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
"""
.. module:: handshaker
.. role:: red

BitDust handshaker() Automat

EVENTS:
    * :red:`ack-received`
    * :red:`ack-timeout`
    * :red:`cache-and-ping`
    * :red:`cancel`
    * :red:`fail-received`
    * :red:`outbox-failed`
    * :red:`ping`
    * :red:`remote-identity-cached`
    * :red:`remote-identity-failed`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 16

#------------------------------------------------------------------------------

from twisted.internet.defer import Deferred

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.lib import strng
from bitdust.lib import packetid

from bitdust.automats import automat

from bitdust.main import settings

from bitdust.contacts import identitycache

from bitdust.crypt import signed

from bitdust.p2p import commands

from bitdust.transport import packet_out
from bitdust.transport import gateway

from bitdust.userid import global_id
from bitdust.userid import my_id

#------------------------------------------------------------------------------

_RunningHandshakers = {}
_KnownChannels = {}

#------------------------------------------------------------------------------


def ping(
    idurl,
    ack_timeout=None,
    cache_timeout=15,
    cache_retries=2,
    ping_retries=2,
    force_cache=False,
    skip_outbox=False,
    keep_alive=True,
    fake_identity=None,
    channel='identity',
    channel_counter=True,
    cancel_running=False,
):
    """
    Doing peer-to-peer ping with acknowledgment and return `Deferred` object to receive result.
    First read remote identity file from `idurl` location.
    Then sending my own identity to remote node and wait for ack.
    If Ack() packet from remote node times out (or another error happened)
    should return failed result in the result `Deferred`.
    """
    global _RunningHandshakers
    remote_idurl = strng.to_bin(idurl)
    if not remote_idurl:
        raise Exception('empty idurl provided')
    result = Deferred()
    pending_results = []
    if remote_idurl in _RunningHandshakers:
        if cancel_running:
            running_info = _RunningHandshakers.pop(remote_idurl)
            inst = running_info.pop('instance')
            pending_results = running_info.pop('results')
            if inst:
                inst.automat('cancel')
        else:
            _RunningHandshakers[remote_idurl]['results'].append(result)
            if _Debug:
                lg.args(_DebugLevel, already_opened=True, idurl=remote_idurl, channel=channel, skip_outbox=skip_outbox)
            return result
    pending_results += [
        result,
    ]
    _RunningHandshakers[remote_idurl] = {
        'instance': None,
        'results': pending_results,
    }
    if _Debug:
        lg.args(_DebugLevel, already_opened=False, idurl=remote_idurl, channel=channel, skip_outbox=skip_outbox, pending_results=len(pending_results))
    h = Handshaker(
        remote_idurl=remote_idurl,
        ack_timeout=ack_timeout,
        cache_timeout=cache_timeout,
        cache_retries=cache_retries,
        ping_retries=ping_retries,
        skip_outbox=skip_outbox,
        keep_alive=keep_alive,
        fake_identity=fake_identity,
        channel=channel,
        channel_counter=channel_counter,
        debug_level=_DebugLevel,
        log_events=_Debug,
        log_transitions=_Debug,
    )
    _RunningHandshakers[remote_idurl]['instance'] = h
    if force_cache:
        h.automat('cache-and-ping')
    else:
        h.automat('ping')
    return result


def is_running(idurl):
    """
    Returns True if some "ping" is currently running towards given user.
    """
    global _RunningHandshakers
    remote_idurl = strng.to_bin(idurl)
    if not remote_idurl:
        return False
    return remote_idurl in _RunningHandshakers


def get_info(idurl):
    """
    If there is a running handshake with given idurl will return its instance.
    """
    global _RunningHandshakers
    remote_idurl = strng.to_bin(idurl)
    if not remote_idurl:
        return None
    return _RunningHandshakers.get(remote_idurl)


def cancel_all():
    global _RunningHandshakers
    for remote_idurl in list(_RunningHandshakers.keys()):
        h = (_RunningHandshakers.get(remote_idurl, {}) or {}).get('instance')
        if h:
            h.event('cancel')


#------------------------------------------------------------------------------


def on_identity_packet_outbox_status(pkt_out, status, error):
    global _RunningHandshakers
    remote_idurl = strng.to_bin(pkt_out.outpacket.RemoteID)
    if remote_idurl in _RunningHandshakers:
        inst = _RunningHandshakers[remote_idurl]['instance']
        if inst:
            if status == 'finished':
                inst.automat('identity-sent', pkt_out)
            else:
                inst.automat('outbox-failed', status=status, error=error)


#------------------------------------------------------------------------------


class Handshaker(automat.Automat):
    """
    This class implements all the functionality of ``handshaker()`` state machine.
    """
    def __init__(
        self, remote_idurl, ack_timeout, cache_timeout, cache_retries, ping_retries, skip_outbox, keep_alive, fake_identity, channel, channel_counter, debug_level=0, log_events=False, log_transitions=False, publish_events=False, **kwargs
    ):
        """
        Builds `handshaker()` state machine.
        """
        global _KnownChannels
        self.remote_idurl = strng.to_bin(remote_idurl)
        self.remote_global_id = global_id.idurl2glob(self.remote_idurl)
        self.ack_timeout = ack_timeout
        if not self.ack_timeout:
            self.ack_timeout = settings.P2PTimeOut()
        self.cache_timeout = cache_timeout
        self.cache_retries = cache_retries
        self.ping_retries = ping_retries
        self.skip_outbox = skip_outbox
        self.keep_alive = keep_alive
        self.fake_identity = fake_identity
        self.channel = channel
        self.channel_counter = channel_counter
        if self.channel not in _KnownChannels:
            _KnownChannels[self.channel] = 0
        _KnownChannels[self.channel] += 1
        super(Handshaker, self).__init__(
            name='handshake_%s%d_%s' % (self.channel.replace('_', ''), _KnownChannels[self.channel], self.remote_global_id), state='AT_STARTUP', debug_level=debug_level, log_events=log_events, log_transitions=log_transitions,
            publish_events=publish_events, **kwargs
        )

    def init(self):
        """
        Method to initialize additional variables and flags
        at creation phase of `handshaker()` machine.
        """
        self.cache_attempts = 0
        self.ping_attempts = 0

    def A(self, event, *args, **kwargs):
        """
        The state machine code, generated using `visio2python <http://bitdust.io/visio2python/>`_ tool.
        """
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'cache-and-ping' or (event == 'ping' and not self.isCached(*args, **kwargs)):
                self.state = 'CACHE'
                self.doInit(*args, **kwargs)
                self.doCacheRemoteIDURL(*args, **kwargs)
            elif event == 'ping' and self.isCached(*args, **kwargs):
                self.state = 'ACK?'
                self.doInit(*args, **kwargs)
                self.doSendMyIdentity(*args, **kwargs)
        #---CACHE---
        elif self.state == 'CACHE':
            if event == 'remote-identity-failed' and not self.isMoreCacheRetries(*args, **kwargs):
                self.state = 'NO_IDENT'
                self.doReportNoIdentity(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'remote-identity-failed' and self.isMoreCacheRetries(*args, **kwargs):
                self.doCacheRemoteIDURL(*args, **kwargs)
            elif event == 'remote-identity-cached':
                self.state = 'ACK?'
                self.doSendMyIdentity(*args, **kwargs)
            elif event == 'cancel':
                self.state = 'NO_IDENT'
                self.doDestroyMe(*args, **kwargs)
        #---ACK?---
        elif self.state == 'ACK?':
            if event == 'ack-received':
                self.state = 'SUCCESS'
                self.doReportSuccess(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'ack-timeout' and not self.isMorePingRetries(*args, **kwargs):
                self.state = 'TIMEOUT'
                self.doReportTimeOut(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'ack-timeout' and self.isMorePingRetries(*args, **kwargs):
                self.doSendMyIdentity(*args, **kwargs)
            elif event == 'fail-received' or event == 'outbox-failed':
                self.state = 'FAILED'
                self.doReportFailed(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'cancel':
                self.state = 'NO_IDENT'
                self.doDestroyMe(*args, **kwargs)
        #---SUCCESS---
        elif self.state == 'SUCCESS':
            pass
        #---TIMEOUT---
        elif self.state == 'TIMEOUT':
            pass
        #---NO_IDENT---
        elif self.state == 'NO_IDENT':
            pass
        #---FAILED---
        elif self.state == 'FAILED':
            pass
        return None

    def isCached(self, *args, **kwargs):
        """
        Condition method.
        """
        return identitycache.HasKey(self.remote_idurl)

    def isMoreCacheRetries(self, *args, **kwargs):
        """
        Condition method.
        """
        return self.cache_attempts <= self.cache_retries

    def isMorePingRetries(self, *args, **kwargs):
        """
        Condition method.
        """
        return self.ping_attempts <= self.ping_retries

    def doInit(self, *args, **kwargs):
        """
        Action method.
        """

    def doCacheRemoteIDURL(self, *args, **kwargs):
        """
        Action method.
        """
        self.cache_attempts += 1
        idcache_defer = identitycache.immediatelyCaching(idurl=strng.to_text(self.remote_idurl), timeout=self.cache_timeout)
        idcache_defer.addCallback(lambda src: self.automat('remote-identity-cached', src))
        idcache_defer.addErrback(lambda err: self.automat('remote-identity-failed', err) and None)

    def doSendMyIdentity(self, *args, **kwargs):
        """
        Action method.
        """
        global _KnownChannels
        self.ping_attempts += 1
        if self.fake_identity:
            identity_object = self.fake_identity
        else:
            identity_object = my_id.getLocalIdentity()
        if not identity_object.Valid():
            raise Exception('can not use invalid identity for ping')
        if self.channel_counter:
            packet_id = '%s:%d:%d:%s' % (self.channel, _KnownChannels[self.channel], self.ping_attempts, packetid.UniqueID())
        else:
            packet_id = '%s:%d:%s' % (self.channel, self.ping_attempts, packetid.UniqueID())
        ping_packet = signed.Packet(
            Command=commands.Identity(),
            OwnerID=my_id.getIDURL(),
            CreatorID=my_id.getIDURL(),
            PacketID=packet_id,
            Payload=strng.to_bin(identity_object.serialize()),
            RemoteID=self.remote_idurl,
        )
        if self.skip_outbox:
            packet_out.create(
                outpacket=ping_packet,
                wide=True,
                response_timeout=self.ack_timeout,
                callbacks={
                    commands.Ack(): lambda response, info: self.automat('ack-received', response=response, info=info),
                    commands.Fail(): lambda response, info: self.automat('fail-received', response=response, info=info),
                    None: lambda pkt_out: self.automat('ack-timeout', pkt_out),
                },
                keep_alive=self.keep_alive,
            )
        else:
            gateway.outbox(
                outpacket=ping_packet,
                wide=True,
                response_timeout=self.ack_timeout,
                callbacks={
                    commands.Ack(): lambda response, info: self.automat('ack-received', response=response, info=info),
                    commands.Fail(): lambda response, info: self.automat('fail-received', response=response, info=info),
                    None: lambda pkt_out: self.automat('ack-timeout', pkt_out),
                },
                keep_alive=self.keep_alive,
            )
        if _Debug:
            lg.args(_DebugLevel, packet_id=packet_id, remote_idurl=self.remote_idurl, ping_attempts=self.ping_attempts)

    def doReportNoIdentity(self, *args, **kwargs):
        """
        Action method.
        """
        global _RunningHandshakers
        lg.warn('failed to cache remote identity %r after %d attempts' % (self.remote_idurl, self.cache_attempts))
        if self.remote_idurl in _RunningHandshakers:
            for result_defer in _RunningHandshakers[self.remote_idurl]['results']:
                if not result_defer.called:
                    result_defer.errback(Exception('failed to cache remote identity %r after %d attempts' % (
                        self.remote_idurl,
                        self.cache_attempts,
                    ), ))

    def doReportFailed(self, *args, **kwargs):
        """
        Action method.
        """
        global _RunningHandshakers
        response = kwargs.get('response')
        info = kwargs.get('info')
        if response and info:
            lg.warn('handshake failed because received Fail() from remote user %r : %r' % (response, info))
            if self.remote_idurl in _RunningHandshakers:
                for result_defer in _RunningHandshakers[self.remote_idurl]['results']:
                    if not result_defer.called:
                        result_defer.errback(Exception('handshake failed because received Fail() packet from remote user'))
            return
        status = kwargs.get('status')
        error = kwargs.get('error')
        if status == 'cancelled':
            lg.warn('handshake was cancelled, my Identity() packet was not sent to %r : %r' % (self.remote_idurl, error))
        else:
            lg.err('handshake failed with status %r, my Identity() packet was not sent to remote user %r : %r' % (status, self.remote_idurl, error))
        if self.remote_idurl in _RunningHandshakers:
            for result_defer in _RunningHandshakers[self.remote_idurl]['results']:
                if not result_defer.called:
                    result_defer.errback(Exception('handshake failed, request packet was not sent to remote node'))

    def doReportTimeOut(self, *args, **kwargs):
        """
        Action method.
        """
        global _RunningHandshakers
        lg.warn('remote node %r did not respond after %d ping attempts' % (self.remote_global_id, self.ping_attempts))
        if self.remote_idurl in _RunningHandshakers:
            for result_defer in _RunningHandshakers[self.remote_idurl]['results']:
                if not result_defer.called:
                    result_defer.errback(Exception('remote node %s did not respond after %d ping attempt(s)' % (
                        self.remote_global_id,
                        self.ping_attempts,
                    ), ))

    def doReportSuccess(self, *args, **kwargs):
        """
        Action method.
        """
        global _RunningHandshakers
        response = kwargs.get('response')
        info = kwargs.get('info')
        if _Debug:
            lg.args(_DebugLevel, channel=self.channel, idurl=self.remote_idurl, response=response, info=info)
        if self.remote_idurl in _RunningHandshakers:
            for result_defer in _RunningHandshakers[self.remote_idurl]['results']:
                if not result_defer.called:
                    result_defer.callback((
                        response,
                        info,
                    ))

    def doDestroyMe(self, *args, **kwargs):
        """
        Remove all references to the state machine object to destroy it.
        """
        global _RunningHandshakers
        if self.remote_idurl in _RunningHandshakers:
            _RunningHandshakers[self.remote_idurl]['instance'] = None
            _RunningHandshakers[self.remote_idurl]['results'] = []
            _RunningHandshakers.pop(self.remote_idurl)
        else:
            lg.warn('did not found registered Handshaker() instance for %r' % self.remote_idurl)
        self.remote_idurl = None
        self.ack_timeout = None
        self.cache_timeout = None
        self.cache_retries = None
        self.ping_retries = None
        self.skip_outbox = None
        self.fake_identity = None
        self.channel = None
        self.channel_counter = None
        self.destroy()



#!/usr/bin/env python
# callback.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (callback.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
..

module:: callback
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 12

#------------------------------------------------------------------------------

from twisted.internet.defer import Deferred

#------------------------------------------------------------------------------

from bitdust.logs import lg

#------------------------------------------------------------------------------

# receiving callbacks
_InterestedParties = {}
_InboxPacketCallbacksList = []
_BeginFileReceivingCallbacksList = []
_FinishFileReceivingCallbacksList = []

# sending callbacks
_OutboxPacketCallbacksList = []
_OutboxPacketFilterCallbacksList = []
_QueueItemStatusCallbacksList = []
_BeginFileSendingCallbacksList = []
_FinishFileSendingCallbacksList = []
_FileSendingFilterCallbacksList = []

#------------------------------------------------------------------------------


class InterestedParty:
    def __init__(self, CallBackFunctionOrDefer, CreatorID, PacketID):
        self.CallBackFunction = CallBackFunctionOrDefer  # function(or Deferred) to call when we see this packet
        self.ComboID = combine_IDs(CreatorID, PacketID)


#------------------------------------------------------------------------------


def interested_parties():
    global _InterestedParties
    return _InterestedParties


#------------------------------------------------------------------------------


def combine_IDs(CreatorID, PacketID):
    return str(CreatorID) + ':' + str(PacketID)


#------------------------------------------------------------------------------


def append_inbox_callback(cb):
    """
    You can add a callback to receive incoming ``packets``. Callback will be
    called with such arguments::

    callback(newpacket, info, status, error_message).
    """
    #     if _Debug:
    #         lg.out(_DebugLevel, 'callback.append_inbox_callback new callback, current callbacks:')
    global _InboxPacketCallbacksList
    if cb not in _InboxPacketCallbacksList:
        _InboxPacketCallbacksList.append(cb)


#     if _Debug:
#         import pprint
#         lg.out(_DebugLevel, '        %s' % pprint.pformat(_InboxPacketCallbacksList))


def insert_inbox_callback(index, cb):
    """
    Same like ``append_inbox_callback(cb)`` but put the callback at the given
    position in the callbacks list. If you put your callback at the top you
    will catch the.

    inbox packet as soon as possible - before other callbacks.
    Callback will be called in a such way:

        callback(newpacket, info, status, error_message).
    """
    global _InboxPacketCallbacksList
    if cb not in _InboxPacketCallbacksList:
        _InboxPacketCallbacksList.insert(index, cb)


def remove_inbox_callback(cb):
    global _InboxPacketCallbacksList
    if cb in _InboxPacketCallbacksList:
        _InboxPacketCallbacksList.remove(cb)


def append_outbox_filter_callback(cb):
    """
    You can add a callback to filter all outgoing traffic. Callback will be
    called with such arguments::

    callback(outpacket, wide, callbacks)
    """
    global _OutboxPacketFilterCallbacksList
    if cb not in _OutboxPacketFilterCallbacksList:
        _OutboxPacketFilterCallbacksList.append(cb)


def insert_outbox_filter_callback(index, cb):
    """
    Same like ``append_outbox_filter_callback(cb)`` but put the callback at the
    given position in the filters list. If you put your callback at the top you
    will catch the outgoing packet as soon as possible - before other callbacks.
    If callback returned True - all other callbacks will be skipped.
    Callback will be called in a such way:

        callback(outpacket, wide, callbacks).
    """
    global _OutboxPacketFilterCallbacksList
    if cb not in _OutboxPacketFilterCallbacksList:
        _OutboxPacketFilterCallbacksList.insert(index, cb)


def remove_outbox_filter_callback(cb):
    global _OutboxPacketFilterCallbacksList
    if cb in _OutboxPacketFilterCallbacksList:
        _OutboxPacketFilterCallbacksList.remove(cb)


def add_outbox_callback(cb):
    """
    You can add a callback to be notified when ``outbox()`` method was called.
    Useful when need to catch that event in third module. Arguments::

    callback(pkt_out)
    """
    global _OutboxPacketCallbacksList
    if cb not in _OutboxPacketCallbacksList:
        _OutboxPacketCallbacksList.append(cb)


def remove_outbox_callback(cb):
    global _OutboxPacketCallbacksList
    if cb in _OutboxPacketCallbacksList:
        _OutboxPacketCallbacksList.remove(cb)


def add_queue_item_status_callback(cb):
    """
    callback(pkt_out, status, error_message)
    """
    global _QueueItemStatusCallbacksList
    if cb not in _QueueItemStatusCallbacksList:
        _QueueItemStatusCallbacksList.append(cb)


def remove_queue_item_status_callback(cb):
    global _QueueItemStatusCallbacksList
    if cb in _QueueItemStatusCallbacksList:
        _QueueItemStatusCallbacksList.remove(cb)


def add_begin_file_sending_callback(cb):
    """
    cb(result_defer, remote_idurl, proto, host, filename, description, pkt_out)
    """
    global _BeginFileSendingCallbacksList
    if cb not in _BeginFileSendingCallbacksList:
        _BeginFileSendingCallbacksList.append(cb)


def remove_begin_file_sending_callback(cb):
    global _BeginFileSendingCallbacksList
    if cb in _BeginFileSendingCallbacksList:
        _BeginFileSendingCallbacksList.remove(cb)


def add_finish_file_sending_callback(cb):
    """
    cb(pkt_out, item, status, size, error_message)
    """
    global _FinishFileSendingCallbacksList
    if cb not in _FinishFileSendingCallbacksList:
        _FinishFileSendingCallbacksList.append(cb)


def remove_finish_file_sending_callback(cb):
    global _FinishFileSendingCallbacksList
    if cb in _FinishFileSendingCallbacksList:
        _FinishFileSendingCallbacksList.remove(cb)


def add_begin_file_receiving_callback(cb):
    """
    cb(pkt_in)
    """
    global _BeginFileReceivingCallbacksList
    if cb not in _BeginFileReceivingCallbacksList:
        _BeginFileReceivingCallbacksList.append(cb)


def remove_begin_file_receiving_callback(cb):
    global _BeginFileReceivingCallbacksList
    if cb in _BeginFileReceivingCallbacksList:
        _BeginFileReceivingCallbacksList.remove(cb)


def add_finish_file_receiving_callback(cb, index=0):
    """
    cb(pkt_in, data)
    """
    global _FinishFileReceivingCallbacksList
    if cb not in _FinishFileReceivingCallbacksList:
        _FinishFileReceivingCallbacksList.insert(index, cb)


def remove_finish_file_receiving_callback(cb):
    global _FinishFileReceivingCallbacksList
    if cb in _FinishFileReceivingCallbacksList:
        _FinishFileReceivingCallbacksList.remove(cb)


def add_file_sending_filter_callback(cb):
    """
    cb(remote_idurl, proto, host, filename, description, pkt_out) : True / False
    """
    global _FileSendingFilterCallbacksList
    if cb not in _FileSendingFilterCallbacksList:
        _FileSendingFilterCallbacksList.append(cb)


def remove_file_sending_filter_callback(cb):
    global _FileSendingFilterCallbacksList
    if cb in _FileSendingFilterCallbacksList:
        _FileSendingFilterCallbacksList.remove(cb)


#------------------------------------------------------------------------------


def run_inbox_callbacks(newpacket, info, status, error_message):
    global _InboxPacketCallbacksList
    if _Debug:
        lg.out(_DebugLevel, 'callback.run_inbox_callbacks for %s from %s' % (newpacket, info))
    handled = False
    for cb in _InboxPacketCallbacksList:
        try:
            _ok = cb(newpacket, info, status, error_message)
            if _ok:
                handled = True
        except:
            lg.exc()
            continue
        if not handled:
            if _Debug:
                lg.out(_DebugLevel, '    passed by %s' % lg.fqn(cb))
            continue
        if _Debug:
            lg.out(_DebugLevel, '    handled by %s' % lg.fqn(cb))
        break
    return handled


def run_outbox_callbacks(pkt_out):
    global _OutboxPacketCallbacksList
    if _Debug:
        lg.out(_DebugLevel, 'callback.run_outbox_callbacks for %s' % pkt_out)
        # lg.out(_DebugLevel, '    %s' % _OutboxPacketCallbacksList)
    handled = False
    for cb in _OutboxPacketCallbacksList:
        try:
            if cb(pkt_out):
                handled = True
        except:
            lg.exc()
        if not handled:
            if _Debug:
                lg.out(_DebugLevel, '    passed by %s' % lg.fqn(cb))
            continue
        if _Debug:
            lg.out(_DebugLevel, '    handled by %s' % lg.fqn(cb))
        break
    return handled


def run_outbox_filter_callbacks(outpacket, wide, callbacks, **kwargs):
    global _OutboxPacketFilterCallbacksList
    for cb in _OutboxPacketFilterCallbacksList:
        try:
            result = cb(outpacket, wide, callbacks, **kwargs)
        except:
            result = None
            lg.exc()
        if result is None:
            # sending was skipped in this filter, use next one
            continue
        if isinstance(result, Deferred):
            # filter was applied to outgoing data, but sending was delayed
            return result
        if result:
            # filter was applied, data was sent, return instance of packet_out.PacketOut
            return result
    lg.warn('no outbox filter was applied to : %s' % outpacket)
    return None


def run_queue_item_status_callbacks(pkt_out, status, error_message):
    global _QueueItemStatusCallbacksList
    handled = False
    for cb in _QueueItemStatusCallbacksList:
        try:
            if cb(pkt_out, status, error_message):
                handled = True
        except:
            lg.exc()
        if handled:
            break
    return handled


def run_file_sending_filter_callbacks(remote_idurl, proto, host, filename, description, pkt_out):
    global _FileSendingFilterCallbacksList
    if _Debug:
        lg.out(_DebugLevel, 'callback.run_file_sending_filter_callbacks %s to %s:%s at %s' % (description, proto, host, remote_idurl))
    one_result = None
    for cb in _FileSendingFilterCallbacksList:
        try:
            one_result = cb(remote_idurl, proto, host, filename, description, pkt_out)
        except:
            lg.exc()
            continue
        if one_result is not None:
            if _Debug:
                lg.out(_DebugLevel, '    filtered by %r : %r' % (cb, one_result))
            break
    return one_result


def run_begin_file_sending_callbacks(result_defer, remote_idurl, proto, host, filename, description, pkt_out):
    global _BeginFileSendingCallbacksList
    if _Debug:
        lg.out(_DebugLevel, 'callback.run_begin_file_sending_callbacks %s to %s:%s at %s' % (description, proto, host, remote_idurl))
    handled = False
    for cb in _BeginFileSendingCallbacksList:
        try:
            if cb(result_defer, remote_idurl, proto, host, filename, description, pkt_out):
                handled = True
        except:
            lg.exc()
        if handled:
            if _Debug:
                lg.out(_DebugLevel, '    handled by %s' % cb)
            break
    return handled


def run_finish_file_sending_callbacks(pkt_out, item, status, size, error_message):
    global _FinishFileSendingCallbacksList
    handled = False
    for cb in _FinishFileSendingCallbacksList:
        try:
            if cb(pkt_out, item, status, size, error_message):
                handled = True
        except:
            lg.exc()
        if handled:
            break
    return handled


def run_begin_file_receiving_callbacks(pkt_in):
    global _BeginFileReceivingCallbacksList
    if _Debug:
        lg.out(_DebugLevel, 'callback.run_begin_file_receiving_callbacks for %s' % pkt_in)
    handled = False
    for cb in _BeginFileReceivingCallbacksList:
        try:
            if cb(pkt_in):
                handled = True
        except:
            lg.exc()
        if handled:
            if _Debug:
                lg.out(_DebugLevel, '    handled by %s' % cb)
            break
    return handled


def run_finish_file_receiving_callbacks(info, data):
    global _FinishFileReceivingCallbacksList
    if _Debug:
        lg.out(_DebugLevel, 'callback.run_finish_file_receiving_callbacks %d bytes : %s' % (len(data), info))
    handled = False
    for cb in _FinishFileReceivingCallbacksList:
        try:
            if cb(info, data):
                handled = True
        except:
            lg.exc()
        if handled:
            if _Debug:
                lg.out(_DebugLevel, '    handled by %s' % cb)
            break
    return handled



#!/usr/bin/python
# nameurl.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (nameurl.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: nameurl.

Here is a methods to work with URL strings.
We assume BitDust URLs are of the form::

    ssh://host.name.com:port/fooobar.xml  (maybe no foobar.xml)
    tcp://123.45.67.89:4321
    udp://veselin@my.identityserver.net:8080/some_folder/veselin

Tried built-in methods ``urlparse()`` and ``urlsplit()``
and they move the location from the second to the 3rd
argument if you have "http" vs "ssh" or "tcp".
This seems like trouble.
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from __future__ import print_function

#------------------------------------------------------------------------------

import six
import six.moves.urllib.parse  # @UnresolvedImport
import six.moves.urllib.parse  # @UnresolvedImport
from six.moves import range  # @UnresolvedImport

#------------------------------------------------------------------------------

import re

#------------------------------------------------------------------------------

from bitdust.lib import strng

#------------------------------------------------------------------------------

legalchars = '#.-_()ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789'
legalset = set(legalchars)

#------------------------------------------------------------------------------


def UrlParse(url):
    """
    Return a tuple of strings from url address : ( proto, host, port, filename )
        import nameurl
        nameurl.UrlParse('http://id.bitdust.io/veselin.xml')
        ('http', 'id.bitdust.io', '', 'veselin.xml')
    """
    if not url or (isinstance(url, six.string_types) and url in [
        'None',
        b'None',
        '',
        b'',
    ]):
        return '', '', '', ''
    url = strng.to_bin(url)
    o = six.moves.urllib.parse.urlparse(url)
    proto = strng.to_bin(o.scheme.strip())
    base = strng.to_bin(o.netloc).lstrip(b' /')
    filename = strng.to_bin(o.path).lstrip(b' /')
    if not base:
        base = strng.to_bin(o.path).lstrip(b' /')
        filename = b''
    if base.find(b'/') < 0:
        if base.find(b':') < 0:
            host = base
            port = b''
        else:
            host, port = base.split(b':', 1)
    else:
        host, tail = base.split(b'/', 1)
        if host.find(b':') < 0:
            port = b''
        else:
            host, port = host.split(b':', 1)
        if not filename:
            filename = tail
    return (
        strng.to_text(proto).strip(),
        strng.to_text(host).strip(),
        strng.to_text(port).strip(),
        strng.to_text(filename).strip(),
    )


def UrlParseFast(url):
    """
    Return a 5-items tuple from url address :
        ( proto, host, port, path, filename )
    """
    url = strng.to_text(url)
    proto, _, tail = url.partition('://')
    head, _, filename = tail.rpartition('/')
    host_port, _, path = head.partition('/')
    host, _, port = host_port.partition(':')
    return proto, host, port, path, filename


def UrlMake(protocol='', machine='', port='', filename='', parts=None):
    """
    Reverse method, create a URL from 4 pieces.
    """
    if parts is not None:
        protocol, machine, port, filename = parts
    url = protocol + '://' + strng.to_text(machine)
    if port:
        url += ':' + strng.to_text(port)
    if filename:
        url += '/' + filename
    return strng.to_bin(url)


def UrlFilename(url):
    """
    Generate a 'safe' filename from URL address.

    This is useful when need to store identity files on disk.
    nameurl.UrlFilename('http://id.bitdust.io/veselin.xml')
    'http###id.bitdust.io#veselin.xml'
    """
    # TODO: switch all that to global ID format
    if not url or (isinstance(url, six.string_types) and url in [
        'None',
        b'None',
        '',
        b'',
    ]):
        return None
    result = strng.to_text(url)
    result = result.replace('://', '###')
    result = result.replace('/', '#')
    result = re.sub('(\:)(\d+)', '(#\g<2>#)', result)
    result = result.lower()
    return result


def FilenameUrl(filename):
    """
    A reverse method for ``UrlFilename()``.
    """
    src = filename.strip().lower()
    if not src.startswith('http###'):
        return None
    src = re.sub('\(#(\d+)#\)', ':\g<1>', src)
    src = 'http://' + src[7:].replace('#', '/')
    return strng.to_bin(src)


def UrlFilenameHTML(url):
    """
    Another method to simplify URL, so you can create a filename from URL
    string.

    nameurl.UrlFilenameHTML('http://id.bitdust.io/veselin.xml')
    'id_bitdust_net_veselin_xml'
    """
    global legalset
    url = strng.to_text(url)
    s = url.replace('http://', '')
    o = ''
    for x in s:
        if x not in legalset or x == '.' or x == ':' or x == '#':
            o += '_'
        else:
            o += x
    return o


#------------------------------------------------------------------------------


def IdContactSplit(contact):
    try:
        return strng.to_text(contact).split('://')
    except:
        return '', ''


def GetName(url):
    """
    Deal with the identities, return a filename (without extension) from URL
    address.

    nameurl.GetName('http://id.bitdust.io/kinggeorge.xml') 'kinggeorge'
    """
    if not url or (isinstance(url, six.string_types) and url in [
        'None',
        b'None',
        '',
        b'',
    ]):
        return ''
    url = strng.to_text(url)
    if not url.endswith('.xml'):
        return url
    return url[url.rfind('/') + 1:-4]  # return url[url.rfind("/")+1:url.rfind(".")]


def GetFileName(url):
    """
    Almost the same, but keeps the file extension.
    """
    if not url or (isinstance(url, six.string_types) and url in [
        'None',
        b'None',
        '',
        b'',
    ]):
        return ''
    url = strng.to_text(url)
    return url[url.rfind('/') + 1:]


def GetHost(url):
    """
    Return host name value from url.
    """
    return UrlParse(url)[1]


#------------------------------------------------------------------------------


def Quote(s):
    """
    A wrapper for built-in method ``urllib.quote()``.
    """
    return six.moves.urllib.parse.quote(s, '')


def UnQuote(s):
    """
    A wrapper for built-in method ``urllib.unquote()``.
    """
    return six.moves.urllib.parse.unquote(s)


#------------------------------------------------------------------------------


def DjangoQuote(s):
    """
    Ensure that primary key values do not confuse the admin URLs by escaping
    any '/', '_' and ':' and similarly problematic characters.

    Similar to urllib.quote, except that the quoting is slightly
    different so that it doesn't get automatically unquoted by the Web
    browser.
    """
    res = list(s)
    for i in range(len(res)):
        c = res[i]
        if c in """:/_#?;@&=+$,"<>%\\""":
            res[i] = '_%02X' % ord(c)
    return ''.join(res)


def DjangoUnQuote(s):
    """
    Undo the effects of quote().

    Based heavily on urllib.unquote().
    """
    mychr = chr
    myatoi = int
    lst = s.split('_')
    res = [lst[0]]
    myappend = res.append
    del lst[0]
    for item in lst:
        if item[1:2]:
            try:
                myappend(mychr(myatoi(item[:2], 16)) + item[2:])
            except ValueError:
                myappend('_' + item)
        else:
            myappend('_' + item)
    return ''.join(res)


#------------------------------------------------------------------------------


def main():
    """
    I used this place for tests.
    """
    print(GetName(str(None)))


if __name__ == '__main__':
    main()



#!/usr/bin/python
# network_config.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (network_config.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: network_config
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

import os

#------------------------------------------------------------------------------

from bitdust.system import bpio
from bitdust.system import local_fs

from bitdust.lib import serialization

from bitdust.main import settings

#------------------------------------------------------------------------------


def find_network_config_file():
    networks_json_path = os.path.join(settings.MetaDataDir(), 'networkconfig')
    if not os.path.isfile(networks_json_path):
        # use hard-coded default_network.json file from the repository root
        networks_json_path = os.path.join(bpio.getExecutableDir(), 'default_network.json')
        if not os.path.isfile(networks_json_path):
            networks_json_path = os.path.join(bpio.getExecutableDir(), '..', 'default_network.json')
    return networks_json_path


def read_network_config_file():
    networks_json_path = find_network_config_file()
    if os.path.isfile(networks_json_path):
        network_info = serialization.BytesToDict(
            local_fs.ReadBinaryFile(networks_json_path),
            keys_to_text=True,
            values_to_text=True,
        )
        if network_info:
            return network_info
    from bitdust.main.default_network import default_network_info  # @UnresolvedImport
    return default_network_info()



#!/usr/bin/python
# ratings.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (ratings.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
#

#------------------------------------------------------------------------------

from __future__ import absolute_import
import os
import sys
import time

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 14

#------------------------------------------------------------------------------

try:
    from twisted.internet import reactor  # @UnresolvedImport
except:
    sys.exit('Error initializing twisted.internet.reactor in ratings.py')

from twisted.internet import task

from bitdust.logs import lg

from bitdust.system import bpio

from bitdust.lib import nameurl

from bitdust.main import settings

from bitdust.contacts import contactsdb

#-------------------------------------------------------------------------------

_LoopCountRatingsTask = None
_IndexMonth = {}
_IndexTotal = {}
_InitDone = False

#-------------------------------------------------------------------------------


def init():
    global _InitDone
    if _InitDone:
        return
    if _Debug:
        lg.out(_DebugLevel, 'ratings.init')
    read_index()
    run()
    _InitDone = True


def shutdown():
    global _InitDone
    if _Debug:
        lg.out(_DebugLevel, 'ratings.shutdown')
    stop()
    _InitDone = False


def run():
    global _LoopCountRatingsTask
    try:
        from bitdust.lib import maths
    except:
        lg.exc()

    stop()
    interval = maths.interval_to_next_hour()
    # debug
    #interval = 5
    reactor.callLater(interval, start)  # @UndefinedVariable
    if _Debug:
        lg.out(_DebugLevel, 'ratings.run will start after %s minutes' % str(interval/60.0))


def start():
    global _LoopCountRatingsTask
    _LoopCountRatingsTask = task.LoopingCall(rate_all_users)
    _LoopCountRatingsTask.start(settings.DefaultAlivePacketTimeOut())
    if _Debug:
        lg.out(_DebugLevel, 'ratings.start will count ratings every %s minutes' % str(settings.DefaultAlivePacketTimeOut()/60.0))


def stop():
    global _LoopCountRatingsTask
    if _LoopCountRatingsTask:
        if _LoopCountRatingsTask.running:
            _LoopCountRatingsTask.stop()
        del _LoopCountRatingsTask
        _LoopCountRatingsTask = None
        if _Debug:
            lg.out(_DebugLevel, 'ratings.stop task finished')


def rating_dir(idurl):
    return os.path.join(settings.RatingsDir(), nameurl.UrlFilename(idurl))


def rating_month_file(idurl, monthstr=None):
    if monthstr is None:
        monthstr = time.strftime('%m%y')
    return os.path.join(rating_dir(idurl), monthstr)


def rating_total_file(idurl):
    return os.path.join(rating_dir(idurl), 'total')


def exist_rating_dir(idurl):
    return bpio._dir_exist(rating_dir(idurl))


def make_rating_dir(idurl):
    bpio._dir_make(rating_dir(idurl))


def read_month_rating_dict(idurl, monthstr=None):
    if monthstr is None:
        monthstr = time.strftime('%m%y')
    return bpio._read_dict(rating_month_file(idurl, monthstr))


def write_month_rating_dict(idurl, rating_dict, monthstr=None):
    if monthstr is None:
        monthstr = time.strftime('%m%y')
    return bpio._write_dict(rating_month_file(idurl, monthstr), rating_dict)


def read_total_rating_dict(idurl):
    return bpio._read_dict(rating_total_file(idurl))


def write_total_rating_dict(idurl, rating_dict):
    return bpio._write_dict(rating_total_file(idurl), rating_dict)


def make_blank_rating_dict():
    return {'all': '0', 'alive': '0'}


def increase_rating(idurl, alive_state):
    if not exist_rating_dir(idurl):
        make_rating_dir(idurl)

    month_rating = read_month_rating_dict(idurl)
    if month_rating is None:
        month_rating = make_blank_rating_dict()
    try:
        mallI = int(month_rating['all'])
        maliveI = int(month_rating['alive'])
    except:
        mallI = 0
        maliveI = 0
    mallI += 1
    if alive_state:
        maliveI += 1
    month_rating['all'] = str(mallI)
    month_rating['alive'] = str(maliveI)
    write_month_rating_dict(idurl, month_rating)

    total_rating = read_total_rating_dict(idurl)
    if total_rating is None:
        total_rating = make_blank_rating_dict()
    try:
        tallI = int(total_rating['all'])
        taliveI = int(total_rating['alive'])
    except:
        tallI = 0
        taliveI = 0
    tallI += 1
    if alive_state:
        taliveI += 1
    total_rating['all'] = str(tallI)
    total_rating['alive'] = str(taliveI)
    write_total_rating_dict(idurl, total_rating)
    return mallI, maliveI, tallI, taliveI


def rate_all_users():
    from bitdust.p2p import online_status
    from bitdust.p2p import p2p_connector
    if not p2p_connector.A():
        lg.warn('ratings update skipped, p2p_connector() is not running')
        return
    if p2p_connector.A().state != 'CONNECTED':
        lg.warn('ratings update skipped, p2p_connector() is CONNECTED')
        return
    if _Debug:
        lg.out(_DebugLevel, 'ratings.rate_all_users')
    monthStr = time.strftime('%B')
    for idurl in contactsdb.contacts_remote(include_all=True):
        if not idurl:
            continue
        isalive = online_status.isOnline(idurl)
        mall, malive, tall, talive = increase_rating(idurl, isalive)
        month_percent = 100.0*float(malive)/float(mall)
        total_percent = 100.0*float(talive)/float(tall)
        if _Debug:
            lg.out(_DebugLevel, '[%6.2f%%: %s/%s] in %s and [%6.2f%%: %s/%s] total - %s' % (month_percent, malive, mall, monthStr, total_percent, talive, tall, nameurl.GetName(idurl)))
    read_index()


def remember_connected_time(idurl):
    if not exist_rating_dir(idurl):
        make_rating_dir(idurl)
    bpio.WriteTextFile(os.path.join(rating_dir(idurl), 'connected'), time.strftime('%d%m%y %H:%M:%S'))


def connected_time(idurl):
    s = bpio.ReadTextFile(os.path.join(rating_dir(idurl), 'connected'))
    if not s:
        return None
    try:
        return time.mktime(time.strptime(s, '%d%m%y %H:%M:%S'))
    except:
        lg.exc()
        return None


def read_all_monthly_ratings(idurl):
    if not exist_rating_dir(idurl):
        return None
    d = {}
    for monthstr in os.listdir(rating_dir(idurl)):
        if monthstr == 'total':
            continue
        if monthstr == 'last':
            continue
        month_rating = read_month_rating_dict(idurl, monthstr)
        if month_rating is None:
            continue
        d[monthstr] = month_rating
    return d


def read_index(monthstr=None):
    global _IndexMonth
    global _IndexTotal
    #out(4, 'ratings.read_index')
    if monthstr is None:
        monthstr = time.strftime('%m%y')
    _IndexMonth.clear()
    _IndexTotal.clear()
    for idurl_filename in os.listdir(settings.RatingsDir()):
        idurl = nameurl.FilenameUrl(idurl_filename)
        if idurl is None:
            continue
        month = read_month_rating_dict(idurl, monthstr)
        total = read_total_rating_dict(idurl)
        _IndexMonth[idurl] = {'all': '0', 'alive': '0'} if month is None else month
        _IndexTotal[idurl] = {'all': '0', 'alive': '0'} if total is None else total
        #out(4, '    [%s]: %s, %s' % (nameurl.GetName(idurl), _IndexMonth[idurl], _IndexTotal[idurl]))


def month(idurl):
    global _IndexMonth
    return _IndexMonth.get(idurl, {'all': '0', 'alive': '0'})


def total(idurl):
    global _IndexTotal
    return _IndexTotal.get(idurl, {'all': '0', 'alive': '0'})


def month_percent(idurl):
    try:
        r = month(idurl)
        return round(100.0*float(r['alive'])/float(r['all']), 2)
    except:
        return 0.0


def total_percent(idurl):
    try:
        r = total(idurl)
        return round(100.0*float(r['alive'])/float(r['all']), 2)
    except:
        return 0.0


#-------------------------------------------------------------------------------


def main():
    pass


if __name__ == '__main__':
    main()



#!/usr/bin/python
# api_rest_http_server.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (api_rest_http_server.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: api_rest_http_server
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from six import PY2

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 12

_APILogFileEnabled = False

#------------------------------------------------------------------------------

import os
import sys
import time

#------------------------------------------------------------------------------

from twisted.internet import reactor  # @UnresolvedImport
from twisted.web.server import Site  # @UnresolvedImport

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.interface import api

from bitdust.lib import strng
from bitdust.lib import jsn
from bitdust.lib import serialization

from bitdust.system import local_fs

from bitdust.main import settings

from bitdust_forks.txrestapi.txrestapi.json_resource import JsonAPIResource, _JsonResource
from bitdust_forks.txrestapi.txrestapi.methods import GET, POST, PUT, DELETE, ALL

#------------------------------------------------------------------------------

_APIListener = None
_APISecret = None

#------------------------------------------------------------------------------


def init(port=None):
    global _APIListener
    global _APILogFileEnabled

    _APILogFileEnabled = settings.config.conf().getBool('logs/api-enabled')

    if _APIListener is not None:
        lg.warn('_APIListener already initialized')
        return

    if not port:
        port = 8180

    read_api_secret()

    current_recursionlimit = None
    if PY2:
        current_recursionlimit = sys.getrecursionlimit()
        sys.setrecursionlimit(2000)

    serve_http(port)
    # serve_https(port)

    if PY2:
        sys.setrecursionlimit(current_recursionlimit)

    lg.out(4, 'api_rest_http_server.init')


def shutdown():
    global _APIListener
    if _APIListener is None:
        lg.warn('_APIListener is None')
        return
    lg.out(4, 'api_rest_http_server.shutdown calling _APIListener.stopListening()')
    _APIListener.stopListening()
    del _APIListener
    _APIListener = None
    lg.out(4, '    _APIListener destroyed')


#------------------------------------------------------------------------------


def read_api_secret():
    global _APISecret
    _APISecret = local_fs.ReadTextFile(settings.APISecretFile())


def serve_https(port):
    global _APIListener
    from bitdust.crypt import certificate

    # server private key
    if os.path.exists(settings.APIServerCertificateKeyFile()):
        server_key_pem = local_fs.ReadBinaryFile(settings.APIServerCertificateKeyFile())
        server_key = certificate.load_private_key(server_key_pem)
    else:
        server_key, server_key_pem = certificate.generate_private_key()
        local_fs.WriteBinaryFile(settings.APIServerCertificateKeyFile(), server_key_pem)
    # server certificate
    if os.path.exists(settings.APIServerCertificateFile()):
        server_cert_pem = local_fs.ReadBinaryFile(settings.APIServerCertificateFile())
    else:
        server_cert_pem = certificate.generate_self_signed_cert(
            hostname=u'localhost',
            ip_addresses=[u'127.0.0.1'],
            server_key=server_key,
        )
        local_fs.WriteBinaryFile(settings.APIServerCertificateFile(), server_cert_pem)
    # client private key
    if os.path.exists(settings.APIClientCertificateKeyFile()):
        client_key_pem = local_fs.ReadBinaryFile(settings.APIClientCertificateKeyFile())
        client_key = certificate.load_private_key(client_key_pem)
    else:
        client_key, client_key_pem = certificate.generate_private_key()
        local_fs.WriteBinaryFile(settings.APIClientCertificateKeyFile(), client_key_pem)
    # client certificate
    if os.path.exists(settings.APIClientCertificateFile()):
        client_cert_pem = local_fs.ReadBinaryFile(settings.APIClientCertificateFile())
        ca_cert_pem = local_fs.ReadBinaryFile(settings.APIServerCertificateFile())
    else:
        ca_cert_pem = local_fs.ReadBinaryFile(settings.APIServerCertificateFile())
        ca_cert = certificate.load_certificate(ca_cert_pem)
        client_cert_pem = certificate.generate_csr_client_cert(
            hostname=u'localhost',
            server_ca_cert=ca_cert,
            server_key=server_key,
            client_key=client_key,
        )
        local_fs.WriteBinaryFile(settings.APIClientCertificateFile(), client_cert_pem)

    try:
        from twisted.internet import ssl  # @UnresolvedImport
        api_resource = BitDustRESTHTTPServer()
        site = BitDustAPISite(api_resource, timeout=None)
        auth = ssl.Certificate.loadPEM(server_cert_pem)
        cert = ssl.PrivateCertificate.loadPEM(server_cert_pem + server_key_pem)
        _APIListener = reactor.listenSSL(port, site, cert.options(auth), interface='127.0.0.1')  # @UndefinedVariable
    except:
        lg.exc()
        os._exit(1)


def serve_http(port):
    global _APIListener
    try:
        api_resource = BitDustRESTHTTPServer()
        site = BitDustAPISite(api_resource, timeout=None)
        _APIListener = reactor.listenTCP(port, site)  # @UndefinedVariable
    except:
        lg.exc()
        os._exit(1)


#------------------------------------------------------------------------------


def _request_arg(request, key, default='', mandatory=False):
    """
    Simplify extracting arguments from url query in request.
    """
    args = request.args or {}
    if key in args:
        values = args.get(key, [default])
        return strng.to_text(values[0]) if values else default
    if strng.to_bin(key) in args:
        values = args.get(strng.to_bin(key), [default])
        return strng.to_text(values[0]) if values else default
    if mandatory:
        raise Exception('mandatory url query argument missed: %s' % key)
    return default


def _request_data(request, mandatory_keys=[], default_value={}):
    """
    Simplify extracting input parameters from request body.
    """
    try:
        input_request_data = request.content.getvalue()
    except:
        input_request_data = request.content.read()
    if not input_request_data:
        if mandatory_keys:
            raise Exception('mandatory json input missed: %s' % mandatory_keys)
        return default_value
    try:
        data = serialization.BytesToDict(
            input_request_data,
            encoding='utf-8',
            keys_to_text=True,
            values_to_text=True,
        )
    except:
        raise Exception('invalid json input')
    for k in mandatory_keys:
        if isinstance(k, tuple):
            found = False
            for f in k:
                if f in data:
                    found = True
            if not found:
                raise Exception('one of mandatory parameters missed: %s' % k)
        else:
            if k not in data:
                raise Exception('one of mandatory parameters missed: %s' % mandatory_keys)
    return data


def _input_value(json_data, keys_list, default_value=None):
    """
    Helper method.
    """
    for key in keys_list:
        if key in json_data:
            return json_data[key]
    return default_value


def _index_or_idurl_or_global_id(
    data,
    index_fields=['index', 'position', 'pos'],
    id_fields=['global_id', 'idurl', 'id'],
):
    """
    Helper method.
    """
    value = _input_value(data, index_fields)
    if value is None:
        value = _input_value(data, id_fields)
    return value


#------------------------------------------------------------------------------


class BitDustAPISite(Site):

    def buildProtocol(self, addr):
        """
        Only accepting connections from local machine!
        """
        if addr.host != '127.0.0.1':
            lg.err('refused connection from remote host: %r' % addr.host)
            return None
        return Site.buildProtocol(self, addr)


class BitDustRESTHTTPServer(JsonAPIResource):

    """
    A set of API method to interract and control locally running BitDust process.
    """

    #------------------------------------------------------------------------------

    def getChild(self, name, request):
        global _APISecret
        if _APISecret:
            api_secret_header = request.getHeader('api_secret')
            if api_secret_header != _APISecret:
                return _JsonResource(
                    dict(status='ERROR', errors=['access denied']),
                    time.time(),
                )
        return JsonAPIResource.getChild(self, name, request)

    def log_request(self, request, callback, args):
        if not callback:
            return None
        uri = request.uri.decode()
        try:
            func_name = callback.im_func.func_name
        except:
            func_name = callback.__name__
        _args = jsn.dict_items_to_text(request.args)
        if not _args:
            _args = _request_data(request)
        else:
            _args = {k: (v[0] if (v and isinstance(v, list)) else v) for k, v in _args.items()}
        if _Debug:
            if uri not in [
                '/v1/event/listen/electron',
                '/v1/network/connected',
                '/v1/process/health',
                '/event/listen/electron/v1',
                '/network/connected/v1',
                '/process/health/v1',
            ] or _DebugLevel > 10:
                lg.out(_DebugLevel, '*** %s:%s  API HTTP  %s(%r)' % (request.method.decode(), uri, func_name, _args))
        if _APILogFileEnabled:
            lg.out(0, '*** %s:%s  HTTP  %s(%r)' % (request.method.decode(), uri, func_name, _args), log_name='api', showtime=True)
        return None

    #------------------------------------------------------------------------------

    @GET('^/p/st$')
    @GET('^/v1/process/stop$')
    @GET('^/process/stop/v1$')
    def process_stop_v1(self, request):
        return api.process_stop(instant=bool(_request_arg(request, 'instant', '1') in ['1', 'true']))

    @GET('^/p/rst$')
    @GET('^/v1/process/restart$')
    @GET('^/process/restart/v1$')
    def process_restart_v1(self, request):
        return api.process_restart()

    @GET('^/p/h$')
    @GET('^/v1/process/health$')
    @GET('^/process/health/v1$')
    def process_health_v1(self, request):
        return api.process_health()

    @GET('^/p/h$')
    @GET('^/v1/process/info$')
    @GET('^/process/info/v1$')
    def process_info_v1(self, request):
        return api.process_info()

    @GET('^/p/d$')
    @GET('^/v1/process/debug$')
    @GET('^/process/debug/v1$')
    def process_debug_v1(self, request):
        return api.process_debug()

    #------------------------------------------------------------------------------

    @POST('^/nw/cr')
    @POST('^/network/create')
    @POST('^/network/create/v1$')
    def network_create_v1(self, request):
        data = _request_data(request, mandatory_keys=['url'])
        return api.network_create(url=data['url'])

    @POST('^/nw/sel')
    @POST('^/network/select')
    @POST('^/network/select/v1$')
    def network_select_v1(self, request):
        data = _request_data(request, mandatory_keys=['name'])
        return api.network_select(name=data['name'])

    @GET('^/nw/con$')
    @GET('^/v1/network/connected$')
    @GET('^/network/connected/v1$')
    def network_connected_v1(self, request):
        return api.network_connected(wait_timeout=int(_request_arg(request, 'wait_timeout', '10')))

    @GET('^/nw/dcon$')
    @GET('^/v1/network/disconnect')
    @GET('^/network/disconnect/v1$')
    def network_disconnect_v1(self, request):
        return api.network_disconnect()

    @GET('^/nw/rcon$')
    @GET('^/v1/network/reconnect$')
    @GET('^/network/reconnect/v1$')
    def network_reconnect_v1(self, request):
        return api.network_reconnect()

    @GET('^/nw/st$')
    @GET('^/v1/network/status$')
    @GET('^/network/status/v1$')
    def network_status_v1(self, request):
        return api.network_status(
            suppliers=bool(_request_arg(request, 'suppliers', '0') in ['1', 'true']),
            customers=bool(_request_arg(request, 'customers', '0') in ['1', 'true']),
            cache=bool(_request_arg(request, 'cache', '0') in ['1', 'true']),
            tcp=bool(_request_arg(request, 'tcp', '0') in ['1', 'true']),
            udp=bool(_request_arg(request, 'udp', '0') in ['1', 'true']),
            proxy=bool(_request_arg(request, 'proxy', '0') in ['1', 'true']),
            dht=bool(_request_arg(request, 'dht', '0') in ['1', 'true']),
        )

    @GET('^/nw/i$')
    @GET('^/v1/network/info$')
    @GET('^/v1/network/details$')
    @GET('^/network/info/v1$')
    @GET('^/network/details/v1$')
    def network_info_v1(self, request):
        return api.network_status(
            suppliers=bool(_request_arg(request, 'suppliers', '1') in ['1', 'true']),
            customers=bool(_request_arg(request, 'customers', '1') in ['1', 'true']),
            cache=bool(_request_arg(request, 'cache', '1') in ['1', 'true']),
            tcp=bool(_request_arg(request, 'tcp', '1') in ['1', 'true']),
            udp=bool(_request_arg(request, 'udp', '1') in ['1', 'true']),
            proxy=bool(_request_arg(request, 'proxy', '1') in ['1', 'true']),
            dht=bool(_request_arg(request, 'dht', '1') in ['1', 'true']),
        )

    @GET('^/nw/cf$')
    @GET('^/v1/network/configuration$')
    @GET('^/network/configuration/v1$')
    def network_configuration_v1(self, request):
        return api.network_configuration()

    @GET('^/nw/stn$')
    @GET('^/v1/network/stun$')
    @GET('^/network/stun/v1$')
    def network_stun_v1(self, request):
        return api.network_stun(
            udp_port=int(_request_arg(request, 'udp_port', 0)) or None,
            dht_port=int(_request_arg(request, 'dht_port', 0)) or None,
        )

    #------------------------------------------------------------------------------

    @GET('^/c/l$')
    @GET('^/v1/config/list$')
    @GET('^/config/list/v1$')
    def config_list_v1(self, request):
        return api.configs_list(
            sort=bool(_request_arg(request, 'sort', '0') in ['1', 'true']),
            include_info=bool(_request_arg(request, 'include_info', '0') in ['1', 'true']),
        )

    @GET('^/c/t$')
    @GET('^/v1/config/tree$')
    @GET('^/config/tree/v1$')
    def configs_tree_v1(self, request):
        return api.configs_tree(include_info=bool(_request_arg(request, 'include_info', '0') in ['1', 'true']))

    @GET('^/c/g/(?P<key1>[^/]+)/(?P<key2>[^/]+)/(?P<key3>[^/]+)/$')
    @GET('^/v1/config/get/(?P<key1>[^/]+)/(?P<key2>[^/]+)/(?P<key3>[^/]+)$')
    @GET('^/config/get/(?P<key1>[^/]+)/(?P<key2>[^/]+)/(?P<key3>[^/]+)/v1$')
    def config_get_l3_v1(self, request, key1, key2, key3):
        return api.config_get(key=(key1 + '/' + key2 + '/' + key3), include_info=bool(_request_arg(request, 'include_info', '0') in ['1', 'true']))

    @GET('^/c/g/(?P<key1>[^/]+)/(?P<key2>[^/]+)/$')
    @GET('^/v1/config/get/(?P<key1>[^/]+)/(?P<key2>[^/]+)$')
    @GET('^/config/get/(?P<key1>[^/]+)/(?P<key2>[^/]+)/v1$')
    def config_get_l2_v1(self, request, key1, key2):
        return api.config_get(key=(key1 + '/' + key2), include_info=bool(_request_arg(request, 'include_info', '0') in ['1', 'true']))

    @GET('^/c/g/(?P<key>[^/]+)/$')
    @GET('^/v1/config/get/(?P<key>[^/]+)$')
    @GET('^/config/get/(?P<key>[^/]+)/v1$')
    def config_get_l1_v1(self, request, key):
        return api.config_get(key=key, include_info=bool(_request_arg(request, 'include_info', '0') in ['1', 'true']))

    @GET('^/c/g$')
    @GET('^/v1/config/get$')
    @GET('^/config/get/v1$')
    def config_get_v1(self, request):
        return api.config_get(key=_request_arg(request, 'key', mandatory=True), include_info=bool(_request_arg(request, 'include_info', '0') in ['1', 'true']))

    @POST('^/c/s/(?P<key1>[^/]+)/(?P<key2>[^/]+)/(?P<key3>[^/]+)/$')
    @POST('^/v1/config/set/(?P<key1>[^/]+)/(?P<key2>[^/]+)/(?P<key3>[^/]+)$')
    @POST('^/config/set/(?P<key1>[^/]+)/(?P<key2>[^/]+)/(?P<key3>[^/]+)/v1$')
    def config_set_l3_v1(self, request, key1, key2, key3):
        data = _request_data(request, mandatory_keys=['value'])
        return api.config_set(key=(key1 + '/' + key2 + '/' + key3), value=data['value'])

    @POST('^/c/s/(?P<key1>[^/]+)/(?P<key2>[^/]+)/$')
    @POST('^/v1/config/set/(?P<key1>[^/]+)/(?P<key2>[^/]+)$')
    @POST('^/config/set/(?P<key1>[^/]+)/(?P<key2>[^/]+)/v1$')
    def config_set_l2_v1(self, request, key1, key2):
        data = _request_data(request, mandatory_keys=['value'])
        return api.config_set(key=(key1 + '/' + key2), value=data['value'])

    @POST('^/c/s/(?P<key>[^/]+)/$')
    @POST('^/v1/config/set/(?P<key>[^/]+)$')
    @POST('^/config/set/(?P<key>[^/]+)/v1$')
    def config_set_l1_v1(self, request, key):
        data = _request_data(request, mandatory_keys=['value'])
        return api.config_set(key=key, value=data['value'])

    @POST('^/c/s$')
    @POST('^/v1/config/set$')
    @POST('^/config/set/v1$')
    def config_set_v1(self, request):
        data = _request_data(request, mandatory_keys=['key', 'value'])
        return api.config_set(key=data['key'], value=data['value'])

    #------------------------------------------------------------------------------

    @GET('^/i/g$')
    @GET('^/v1/identity/get$')
    @GET('^/identity/get/v1$')
    def identity_get_v1(self, request):
        return api.identity_get(include_xml_source=bool(_request_arg(request, 'xml_source', '0') in ['1', 'true']))

    @POST('^/i/c$')
    @POST('^/v1/identity/create$')
    @POST('^/identity/create/v1$')
    def identity_create_v1(self, request):
        data = _request_data(request, mandatory_keys=['username'])
        return api.identity_create(username=data['username'], join_network=bool(data.get('join_network', '0') in ['1', 'true']))

    @POST('^/i/b')
    @POST('^/identity/backup$')
    @POST('^/identity/backup/v1$')
    def identity_backup_v1(self, request):
        data = _request_data(request, mandatory_keys=['destination_filepath'])
        return api.identity_backup(destination_filepath=data['destination_filepath'])

    @POST('^/i/r$')
    @POST('^/v1/identity/recover$')
    @POST('^/identity/recover/v1$')
    def identity_recover_v1(self, request):
        data = _request_data(request)
        private_key_source = data.get('private_key_source')
        if not private_key_source:
            private_key_local_file = data.get('private_key_local_file')
            if private_key_local_file:
                from bitdust.system import bpio
                private_key_source = bpio.ReadTextFile(bpio.portablePath(private_key_local_file))
        return api.identity_recover(
            private_key_source=private_key_source,
            known_idurl=data.get('known_idurl'),
            join_network=bool(data.get('join_network', '0') in ['1', 'true']),
        )

    @DELETE('^/i/d$')
    @DELETE('^/v1/identity/erase$')
    @DELETE('^/identity/erase/v1$')
    def identity_erase_v1(self, request):
        data = _request_data(request)
        return api.identity_erase(erase_private_key=data.get('erase_private_key', False))

    @PUT('^/i/rot$')
    @PUT('^/v1/identity/rotate$')
    @PUT('^/identity/rotate/v1$')
    def identity_rotate_v1(self, request):
        return api.identity_rotate()

    @PUT('^/i/h$')
    @PUT('^/v1/identity/heal$')
    @PUT('^/identity/heal/v1$')
    def identity_heal_v1(self, request):
        return api.ERROR('not implemented yet')

    @GET('^/i/ch/l$')
    @GET('^/v1/identity/cache/list$')
    @GET('^/identity/cache/list/v1$')
    def identity_list_v1(self, request):
        return api.identity_cache_list()

    #------------------------------------------------------------------------------

    @GET('^/k/l$')
    @GET('^/v1/key/list$')
    @GET('^/key/list/v1$')
    def key_list_v1(self, request):
        return api.keys_list(
            sort=bool(_request_arg(request, 'sort', '0') in ['1', 'true']),
            include_private=bool(_request_arg(request, 'include_private', '0') in ['1', 'true']),
        )

    @GET('^/k/g$')
    @GET('^/v1/key/get$')
    @GET('^/key/get/v1$')
    def key_get_v1(self, request):
        return api.key_get(
            key_id=_request_arg(request, 'key_id', mandatory=True),
            include_private=bool(_request_arg(request, 'include_private', '0') in ['1', 'true']),
            include_signature=bool(_request_arg(request, 'include_signature', '0') in ['1', 'true']),
            generate_signature=bool(_request_arg(request, 'generate_signature', '0') in ['1', 'true']),
        )

    @POST('^/k/c$')
    @POST('^/v1/key/create$')
    @POST('^/key/create/v1$')
    def key_create_v1(self, request):
        data = _request_data(request, mandatory_keys=['alias'])
        return api.key_create(
            key_alias=data['alias'],
            key_size=int(data['key_size']) if 'key_size' in data else None,
            label=data.get('label', ''),
            active=bool(data.get('active', '1') in ['1', 'true']),
            include_private=bool(data.get('include_private', '0') in ['1', 'true']),
        )

    @POST('^/k/lb$')
    @POST('^/v1/key/label$')
    @POST('^/key/label/v1$')
    def key_label_v1(self, request):
        data = _request_data(request, mandatory_keys=['label', 'key_id'])
        return api.key_label(key_id=data['key_id'], label=data['label'])

    @POST('^/k/st$')
    @POST('^/v1/key/state$')
    @POST('^/key/state/v1$')
    def key_state_v1(self, request):
        data = _request_data(request, mandatory_keys=['active', 'key_id'])
        return api.key_label(key_id=data['key_id'], active=bool(data.get('active', '1') in ['1', 'true']))

    @DELETE('^/k/d$')
    @DELETE('^/v1/key/erase$')
    @DELETE('^/key/erase/v1$')
    def key_erase_v1(self, request):
        data = _request_data(request, mandatory_keys=['key_id'])
        return api.key_erase(key_id=data['key_id'])

    @PUT('^/k/s$')
    @PUT('^/v1/key/share$')
    @PUT('^/key/share/v1$')
    def key_share_v1(self, request):
        data = _request_data(request, mandatory_keys=['key_id', 'trusted_user_id'])
        return api.key_share(
            key_id=data['key_id'],
            trusted_user_id=data['trusted_user_id'],
            include_private=bool(data.get('include_private', '0') in ['1', 'true']),
            include_signature=bool(data.get('include_signature', '0') in ['1', 'true']),
        )

    @POST('^/k/a$')
    @POST('^/v1/key/audit$')
    @POST('^/key/audit/v1$')
    def key_audit_v1(self, request):
        data = _request_data(request, mandatory_keys=['key_id', 'untrusted_user_id'])
        return api.key_audit(
            key_id=data['key_id'],
            untrusted_user_id_id=data['untrusted_user_id'],
            is_private=bool(data.get('is_private', '0') in ['1', 'true']),
        )

    #------------------------------------------------------------------------------

    @GET('^/f/s$')
    @GET('^/v1/file/sync$')
    @GET('^/file/sync/v1$')
    def file_sync_v1(self, request):
        return api.files_sync(force=bool(_request_arg(request, 'force', '0') in ['1', 'true']), )

    @GET('^/f/l$')
    @GET('^/v1/file/list$')
    @GET('^/file/list/v1$')
    def file_list_v1(self, request):
        return api.files_list(
            remote_path=_request_arg(request, 'remote_path', None),
            key_id=_request_arg(request, 'key_id', None),
            recursive=bool(_request_arg(request, 'recursive', '0') in ['1', 'true']),
            all_customers=bool(_request_arg(request, 'all_customers', '0') in ['1', 'true']),
            include_uploads=bool(_request_arg(request, 'uploads', '0') in ['1', 'true']),
            include_downloads=bool(_request_arg(request, 'downloads', '0') in ['1', 'true']),
        )

    @GET('^/f/l/a$')
    @GET('^/v1/file/list/all$')
    @GET('^/file/list/all/v1$')
    def file_list_all_v1(self, request):
        return api.files_list(all_customers=True, include_uploads=True, include_downloads=True)

    @GET('^/f/e$')
    @GET('^/v1/file/exists$')
    @GET('^/file/exists/v1$')
    def file_exists_v1(self, request):
        return api.file_info(remote_path=_request_arg(request, 'remote_path', mandatory=True))

    @GET('^/f/i$')
    @GET('^/v1/file/info$')
    @GET('^/file/info/v1$')
    def file_info_v1(self, request):
        return api.file_info(
            remote_path=_request_arg(request, 'remote_path', mandatory=True),
            include_uploads=bool(_request_arg(request, 'uploads', '1') in ['1', 'true']),
            include_downloads=bool(_request_arg(request, 'downloads', '1') in ['1', 'true']),
        )

    @POST('^/f/c$')
    @POST('^/v1/file/create$')
    @POST('^/file/create/v1$')
    def file_create_v1(self, request):
        data = _request_data(request, mandatory_keys=['remote_path'])
        return api.file_create(
            remote_path=data['remote_path'],
            as_folder=bool(data.get('as_folder', '0') in ['1', 'true']),
        )

    @DELETE('^/f/d$')
    @DELETE('^/v1/file/delete$')
    @DELETE('^/file/delete/v1$')
    def file_delete_v1(self, request):
        data = _request_data(request, mandatory_keys=['remote_path'])
        return api.file_delete(remote_path=data['remote_path'])

    @GET('^/f/u/l$')
    @GET('^/v1/file/upload$')
    @GET('^/file/upload/v1$')
    def files_uploads_v1(self, request):
        return api.files_uploads(
            include_running=bool(_request_arg(request, 'running', '1') in ['1', 'true']),
            include_pending=bool(_request_arg(request, 'pending', '1') in ['1', 'true']),
        )

    @POST('^/f/u/o$')
    @POST('^/v1/file/upload/start$')
    @POST('^/file/upload/start/v1$')
    def file_upload_start_v1(self, request):
        data = _request_data(request, mandatory_keys=['local_path', 'remote_path'])
        return api.file_upload_start(
            local_path=data['local_path'],
            remote_path=data['remote_path'],
            wait_result=bool(data.get('wait_result', '0') in ['1', 'true']),
            publish_events=bool(data.get('publish_events', '0') in ['1', 'true']),
        )

    @POST('^/f/u/c$')
    @POST('^/v1/file/upload/stop$')
    @POST('^/file/upload/stop/v1$')
    def file_upload_stop_v1(self, request):
        data = _request_data(request, mandatory_keys=['remote_path'])
        return api.file_upload_stop(remote_path=data['remote_path'])

    @GET('^/f/d/l$')
    @GET('^/v1/file/download$')
    @GET('^/file/download/v1$')
    def files_downloads_v1(self, request):
        return api.files_downloads()

    @POST('^/f/d/o$')
    @POST('^/v1/file/download/start$')
    @POST('^/file/download/start/v1$')
    def file_download_start_v1(self, request):
        data = _request_data(request, mandatory_keys=['remote_path'])
        return api.file_download_start(
            remote_path=data['remote_path'],
            destination_path=data.get('destination_folder', None),
            wait_result=bool(data.get('wait_result', '0') in ['1', 'true']),
            publish_events=bool(data.get('publish_events', '0') in ['1', 'true']),
        )

    @POST('^/f/d/c$')
    @POST('^/v1/file/download/stop$')
    @POST('^/file/download/stop/v1$')
    def file_download_stop_v1(self, request):
        data = _request_data(request, mandatory_keys=['remote_path'])
        return api.file_download_stop(remote_path=data['remote_path'])

    @GET('^/f/x$')
    @GET('^/v1/file/explore$')
    @GET('^/file/explore/v1$')
    def file_explore_v1(self, request):
        return api.file_explore(local_path=_request_arg(request, 'local_path', mandatory=True))

    #------------------------------------------------------------------------------

    @GET('^/sh/l$')
    @GET('^/v1/share/list$')
    @GET('^/share/list/v1$')
    def share_list_v1(self, request):
        return api.shares_list(
            only_active=bool(_request_arg(request, 'active', '0') in ['1', 'true']),
            include_mine=bool(_request_arg(request, 'mine', '1') in ['1', 'true']),
            include_granted=bool(_request_arg(request, 'granted', '1') in ['1', 'true']),
        )

    @GET('^/sh/i$')
    @GET('^/v1/share/info$')
    @GET('^/share/info/v1$')
    def share_info_v1(self, request):
        return api.share_info(key_id=_request_arg(request, 'key_id', mandatory=True))

    @POST('^/sh/c$')
    @POST('^/v1/share/create$')
    @POST('^/share/create/v1$')
    def share_create_v1(self, request):
        data = _request_data(request)
        return api.share_create(
            owner_id=data.get('owner_id', None),
            key_size=int(data['key_size']) if 'key_size' in data else None,
            label=data.get('label', ''),
            active=bool(data.get('active', '1') in ['1', 'true']),
        )

    @DELETE('^/sh/d$')
    @DELETE('^/v1/share/delete$')
    @DELETE('^/share/delete/v1$')
    def share_delete_v1(self, request):
        data = _request_data(request, mandatory_keys=['key_id'])
        return api.share_delete(key_id=data['key_id'])

    @PUT('^/sh/g$')
    @PUT('^/v1/share/grant$')
    @PUT('^/share/grant/v1$')
    def share_grant_v1(self, request):
        data = _request_data(request, mandatory_keys=[
            ('trusted_user_id', 'trusted_global_id', 'trusted_idurl', 'trusted_id'),
            'key_id',
        ])
        return api.share_grant(
            key_id=data['key_id'],
            trusted_user_id=data.get('trusted_user_id') or data.get('trusted_global_id') or data.get('trusted_idurl') or data.get('trusted_id'),
            timeout=data.get('timeout', 30),
            publish_events=bool(data.get('publish_events', '0') in ['1', 'true']),
        )

    @POST('^/sh/o$')
    @POST('^/v1/share/open$')
    @POST('^/share/open/v1$')
    def share_open_v1(self, request):
        data = _request_data(request, mandatory_keys=['key_id'])
        return api.share_open(
            key_id=data['key_id'],
            publish_events=bool(data.get('publish_events', '0') in ['1', 'true']),
        )

    @DELETE('^/sh/cl$')
    @DELETE('^/v1/share/close$')
    @DELETE('^/share/close/v1$')
    def share_close_v1(self, request):
        data = _request_data(request, mandatory_keys=['key_id'])
        return api.share_close(key_id=data['key_id'])

    @GET('^/sh/h$')
    @GET('^/v1/share/history$')
    @GET('^/share/history/v1$')
    def share_history_v1(self, request):
        return api.share_history()

    #------------------------------------------------------------------------------

    @GET('^/gr/l$')
    @GET('^/v1/group/list$')
    @GET('^/group/list/v1$')
    def group_list_v1(self, request):
        return api.groups_list()

    @POST('^/gr/c$')
    @POST('^/v1/group/create$')
    @POST('^/group/create/v1$')
    def group_create_v1(self, request):
        data = _request_data(request)
        return api.group_create(
            creator_id=data.get('creator_id', None),
            key_size=int(data['key_size']) if 'key_size' in data else None,
            label=data.get('label', ''),
            timeout=data.get('timeout', 20),
        )

    @GET('^/gr/i$')
    @GET('^/v1/group/info$')
    @GET('^/group/info/v1$')
    def group_info_v1(self, request):
        return api.group_info(group_key_id=_request_arg(request, 'group_key_id'))

    @GET('^/gr/dht$')
    @GET('^/v1/group/info/dht$')
    @GET('^/group/info/dht/v1$')
    def group_info_dht_v1(self, request):
        return api.group_info_dht(group_creator_id=_request_arg(request, 'group_creator_id') or _request_arg(request, 'group_creator_idurl') or _request_arg(request, 'id'))

    @POST('^/gr/j$')
    @POST('^/v1/group/join$')
    @POST('^/group/join/v1$')
    def group_join_v1(self, request):
        data = _request_data(request, mandatory_keys=['group_key_id'])
        return api.group_join(
            group_key_id=data['group_key_id'],
            publish_events=bool(data.get('publish_events', '0') in ['1', 'true']),
            use_dht_cache=bool(data.get('use_dht_cache', '0') in ['1', 'true']),
            wait_result=bool(data.get('wait_result', '1') in ['1', 'true']),
        )

    @DELETE('^/gr/lv$')
    @DELETE('^/v1/group/leave$')
    @DELETE('^/group/leave/v1$')
    def group_leave_v1(self, request):
        data = _request_data(request, mandatory_keys=['group_key_id'])
        return api.group_leave(
            group_key_id=data['group_key_id'],
            erase_key=data.get('erase_key', False),
        )

    @PUT('^/gr/r$')
    @PUT('^/v1/group/reconnect$')
    @PUT('^/group/reconnect/v1$')
    def group_reconnect_v1(self, request):
        data = _request_data(request, mandatory_keys=['group_key_id'])
        return api.group_reconnect(
            group_key_id=data['group_key_id'],
            use_dht_cache=bool(data.get('use_dht_cache', '0') in ['1', 'true']),
        )

    @PUT('^/gr/sh$')
    @PUT('^/v1/group/share$')
    @PUT('^/group/share/v1$')
    def group_share_v1(self, request):
        data = _request_data(request, mandatory_keys=[
            ('trusted_user_id', 'trusted_global_id', 'trusted_idurl', 'trusted_id'),
            'group_key_id',
        ])
        return api.group_share(
            group_key_id=data['group_key_id'],
            trusted_user_id=data.get('trusted_user_id') or data.get('trusted_global_id') or data.get('trusted_idurl') or data.get('trusted_id'),
            timeout=data.get('timeout', 45),
            publish_events=bool(data.get('publish_events', '0') in ['1', 'true']),
        )

    #------------------------------------------------------------------------------

    @GET('^/fr/l$')
    @GET('^/v1/friend/list$')
    @GET('^/friend/list/v1$')
    def friend_list_v1(self, request):
        return api.friends_list()

    @POST('^/fr/a$')
    @POST('^/v1/friend/add$')
    @POST('^/friend/add/v1$')
    def friend_add_v1(self, request):
        data = _request_data(request, mandatory_keys=[('trusted_user_id', 'idurl', 'global_id', 'id')])
        return api.friend_add(
            trusted_user_id=data.get('trusted_user_id') or data.get('global_id') or data.get('idurl') or data.get('id'),
            alias=data.get('alias', ''),
        )

    @DELETE('^/fr/d$')
    @DELETE('^/v1/friend/remove$')
    @DELETE('^/friend/remove/v1$')
    def friend_remove_v1(self, request):
        data = _request_data(request, mandatory_keys=[('user_id', 'idurl', 'global_id', 'id')])
        return api.friend_remove(user_id=data.get('user_id') or data.get('global_id') or data.get('idurl') or data.get('id'))

    #------------------------------------------------------------------------------

    @POST('^/us/png$')
    @POST('^/v1/user/ping$')
    @POST('^/user/ping/v1$')
    def user_ping_v1(self, request):
        data = _request_data(request, mandatory_keys=[('user_id', 'idurl', 'global_id', 'id')])
        return api.user_ping(
            user_id=data.get('user_id') or data.get('global_id') or data.get('idurl') or data.get('id'),
            timeout=data.get('timeout', 15),
            retries=data.get('retries', 2),
        )

    @GET('^/us/png$')
    @GET('^/v1/user/ping$')
    @GET('^/user/ping/v1$')
    def user_ping_get_v1(self, request):
        return api.user_ping(
            user_id=_request_arg(request, 'user_id') or _request_arg(request, 'global_id') or _request_arg(request, 'idurl') or _request_arg(request, 'id'),
            timeout=_request_arg(request, 'timeout', 15),
            retries=_request_arg(request, 'retries', 2),
        )

    @GET('^/us/st$')
    @GET('^/v1/user/status$')
    @GET('^/user/status/v1$')
    def user_status_v1(self, request):
        return api.user_status(user_id=_request_arg(request, 'global_id') or _request_arg(request, 'idurl') or _request_arg(request, 'id'))

    @GET('^/us/st/c$')
    @GET('^/v1/user/status/check$')
    @GET('^/user/status/check/v1$')
    def user_status_check_v1(self, request):
        return api.user_status_check(
            user_id=_request_arg(request, 'user_id') or _request_arg(request, 'global_id') or _request_arg(request, 'idurl') or _request_arg(request, 'id'),
            timeout=_request_arg(request, 'timeout', 15),
        )

    @GET('^/us/s/(?P<nickname>[^/]+)/$')
    @GET('^/v1/user/search/(?P<nickname>[^/]+)$')
    @GET('^/user/search/(?P<nickname>[^/]+)/v1$')
    def user_search_v1(self, request, nickname):
        return api.user_search(nickname, attempts=int(_request_arg(request, 'attempts', 1)))

    @GET('^/us/s$')
    @GET('^/v1/user/search$')
    @GET('^/user/search/v1$')
    def user_search_arg_v1(self, request):
        return api.user_search(
            nickname=_request_arg(request, 'nickname', mandatory=True),
            attempts=int(_request_arg(request, 'attempts', 1)),
        )

    @GET('^/us/o/(?P<nickname>[^/]+)/$')
    @GET('^/v1/user/observe/(?P<nickname>[^/]+)$')
    @GET('^/user/observe/(?P<nickname>[^/]+)/v1$')
    def user_observe_v1(self, request, nickname):
        return api.user_observe(
            nickname=nickname,
            attempts=int(_request_arg(request, 'attempts', 3)),
        )

    @GET('^/us/o$')
    @GET('^/v1/user/observe$')
    @GET('^/user/observe/v1$')
    def user_observe_arg_v1(self, request):
        return api.user_observe(
            nickname=_request_arg(request, 'nickname', mandatory=True),
            attempts=int(_request_arg(request, 'attempts', 3)),
        )

    #------------------------------------------------------------------------------

    @GET('^/msg/h$')
    @GET('^/v1/message/history$')
    @GET('^/message/history/v1$')
    def message_history_v1(self, request):
        return api.message_history(
            recipient_id=_request_arg(request, 'id', None, True),
            sender_id=_request_arg(request, 'sender_id', None, False),
            message_type=_request_arg(request, 'message_type', 'private_message'),
            offset=int(_request_arg(request, 'offset', '0')),
            limit=int(_request_arg(request, 'limit', '100')),
        )

    @GET('^/msg/c$')
    @GET('^/v1/message/conversation$')
    @GET('^/message/conversation/v1$')
    def message_conversation_v1(self, request):
        return api.message_conversations_list(
            message_types=list(filter(None,
                                      _request_arg(request, 'message_types', '').split(','))),
            offset=int(_request_arg(request, 'offset', '0')),
            limit=int(_request_arg(request, 'limit', '100')),
        )

    @GET('^/msg/r/(?P<consumer_callback_id>[^/]+)/$')
    @GET('^/v1/message/receive/(?P<consumer_callback_id>[^/]+)$')
    @GET('^/message/receive/(?P<consumer_callback_id>[^/]+)/v1$')
    def message_receive_v1(self, request, consumer_callback_id):
        return api.message_receive(
            consumer_callback_id=consumer_callback_id,
            direction=_request_arg(request, 'direction', 'incoming'),
            message_types=_request_arg(request, 'message_types', 'private_message,group_message'),
            polling_timeout=int(_request_arg(request, 'polling_timeout', 60, False)),
        )

    @POST('^/msg/s$')
    @POST('^/v1/message/send$')
    @POST('^/message/send/v1$')
    def message_send_v1(self, request):
        data = _request_data(request, mandatory_keys=[
            ('recipient_id', 'idurl', 'global_id', 'id'),
            'data',
        ])
        return api.message_send(
            recipient_id=data.get('recipient_id') or data.get('global_id') or data.get('idurl') or data.get('id'),
            data=data['data'],
            ping_timeout=data.get('ping_timeout', 30),
            message_ack_timeout=data.get('message_ack_timeout', 15),
        )

    @POST('^/msg/sg$')
    @POST('^/v1/message/send/group$')
    @POST('^/message/send/group/v1$')
    def message_send_group_v1(self, request):
        data = _request_data(request, mandatory_keys=['group_key_id', 'data'])
        return api.message_send_group(group_key_id=data.get('group_key_id'), data=data['data'])

    #------------------------------------------------------------------------------

    @GET('^/su/l$')
    @GET('^/v1/supplier/list$')
    @GET('^/supplier/list/v1$')
    def supplier_list_v1(self, request):
        return api.suppliers_list(
            customer_id=_request_arg(request, 'customer_id') or _request_arg(request, 'customer_idurl') or _request_arg(request, 'id'),
            verbose=bool(_request_arg(request, 'verbose', '0') in ['1', 'true']),
        )

    @POST('^/su/c$')
    @POST('^/v1/supplier/change')
    @POST('^/supplier/change/v1$')
    def supplier_change_v1(self, request):
        data = _request_data(request)
        return api.supplier_change(
            position=_input_value(data, ['position', 'pos', 'index'], None),
            supplier_id=data.get('supplier_id') or data.get('supplier_idurl') or data.get('supplier_glob_id'),
            new_supplier_id=data.get('new_global_id') or data.get('new_idurl') or data.get('new_supplier_id'),
        )

    @DELETE('^/su/r$')
    @DELETE('^/v1/supplier/replace$')
    @DELETE('^/supplier/replace/v1$')
    def supplier_replace_v1(self, request):
        data = _request_data(request)
        return api.supplier_change(
            position=_input_value(data, ['position', 'pos', 'index'], None),
            supplier_id=data.get('supplier_id') or data.get('supplier_idurl') or data.get('supplier_glob_id'),
            new_supplier_id=None,
        )

    @PUT('^/su/sw$')
    @PUT('^/v1/supplier/switch$')
    @PUT('^/supplier/switch/v1$')
    def supplier_switch_v1(self, request):
        data = _request_data(request, mandatory_keys=[('new_idurl', 'new_global_id', 'new_supplier_id')])
        return api.supplier_change(
            position=_input_value(data, ['position', 'pos', 'index'], None),
            supplier_id=data.get('supplier_id') or data.get('supplier_idurl') or data.get('supplier_glob_id'),
            new_supplier_id=data.get('new_global_id') or data.get('new_idurl') or data.get('new_supplier_id'),
        )

    @POST('^/su/png$')
    @POST('^/v1/supplier/ping$')
    @POST('^/supplier/ping/v1$')
    def supplier_ping_v1(self, request):
        return api.suppliers_ping()

    @GET('^/su/dht$')
    @GET('^/v1/supplier/list/dht$')
    @GET('^/supplier/list/dht/v1$')
    def suppliers_list_dht(self, request):
        return api.suppliers_list_dht(customer_id=_request_arg(request, 'customer_id') or _request_arg(request, 'customer_idurl') or _request_arg(request, 'id'))

    #------------------------------------------------------------------------------

    @GET('^/cu/l$')
    @GET('^/v1/customer/list$')
    @GET('^/customer/list/v1$')
    def customer_list_v1(self, request):
        return api.customers_list()

    @DELETE('^/cu/d$')
    @DELETE('^/v1/customer/reject$')
    @DELETE('^/customer/reject/v1$')
    def customer_reject_v1(self, request):
        data = _request_data(request, mandatory_keys=[('customer_id', 'idurl', 'global_id', 'id')])
        return api.customer_reject(
            customer_id=data.get('customer_id') or data.get('global_id') or data.get('idurl') or data.get('id'),
            erase_customer_key=bool(_request_arg(request, 'erase_customer_key', '1') in ['1', 'true']),
        )

    @POST('^/cu/png$')
    @POST('^/v1/customer/ping$')
    @POST('^/customer/ping/v1$')
    def customer_ping_v1(self, request):
        return api.customers_ping()

    #------------------------------------------------------------------------------

    @GET('^/sp/d$')
    @GET('^/v1/space/donated$')
    @GET('^/space/donated/v1$')
    def space_donated_v1(self, request):
        return api.space_donated()

    @GET('^/sp/c$')
    @GET('^/v1/space/consumed$')
    @GET('^/space/consumed/v1$')
    def space_consumed_v1(self, request):
        return api.space_consumed()

    @GET('^/sp/l$')
    @GET('^/v1/space/local$')
    @GET('^/space/local/v1$')
    def space_local_v1(self, request):
        return api.space_local()

    #------------------------------------------------------------------------------

    @GET('^/svc/l$')
    @GET('^/v1/service/list$')
    @GET('^/service/list/v1$')
    def service_list_v1(self, request):
        return api.services_list(with_configs=bool(_request_arg(request, 'with_configs', '0') in ['1', 'true']))

    @GET('^/svc/i/(?P<service_name>[^/]+)/$')
    @GET('^/v1/service/info/(?P<service_name>[^/]+)$')
    @GET('^/service/info/(?P<service_name>[^/]+)/v1$')
    def service_info_v1(self, request, service_name):
        return api.service_info(service_name)

    @POST('^/svc/o/(?P<service_name>[^/]+)/$')
    @POST('^/v1/service/start/(?P<service_name>[^/]+)$')
    @POST('^/service/start/(?P<service_name>[^/]+)/v1$')
    def service_start_v1(self, request, service_name):
        return api.service_start(service_name)

    @POST('^/svc/c/(?P<service_name>[^/]+)/$')
    @POST('^/v1/service/stop/(?P<service_name>[^/]+)$')
    @POST('^/service/stop/(?P<service_name>[^/]+)/v1$')
    def service_stop_v1(self, request, service_name):
        return api.service_stop(service_name)

    @POST('^/svc/r/(?P<service_name>[^/]+)/$')
    @POST('^/v1/service/restart/(?P<service_name>[^/]+)$')
    @POST('^/service/restart/(?P<service_name>[^/]+)/v1$')
    def service_restart_v1(self, request, service_name):
        return api.service_restart(
            service_name=service_name,
            wait_timeout=_request_data(request).get('wait_timeout', 15),
        )

    @GET('^/svc/h/(?P<service_name>[^/]+)/$')
    @GET('^/v1/service/health/(?P<service_name>[^/]+)$')
    @GET('^/service/health/(?P<service_name>[^/]+)/v1$')
    def service_health_v1(self, request, service_name):
        return api.service_health(service_name)

    #------------------------------------------------------------------------------

    @GET('^/pkt/l$')
    @GET('^/v1/packet/list$')
    @GET('^/packet/list/v1$')
    def packet_list_v1(self, request):
        return api.packets_list()

    @GET('^/pkt/i$')
    @GET('^/v1/packet/info$')
    @GET('^/v1/packet/stats$')
    @GET('^/packet/info/v1$')
    @GET('^/packet/stats/v1$')
    def packet_stats_v1(self, request):
        return api.packets_stats()

    #------------------------------------------------------------------------------

    @GET('^/tr/l$')
    @GET('^/v1/transfer/list$')
    @GET('^/transfer/list/v1$')
    def transfer_list_v1(self, request):
        return api.transfers_list()

    @GET('^/con/l$')
    @GET('^/v1/connection/list$')
    @GET('^/connection/list/v1$')
    def connection_list_v1(self, request):
        return api.connections_list(protocols=map(strng.to_text, filter(None, _request_arg(request, 'protocols', '').strip().lower().split(','))) or None)

    @GET('^/str/l$')
    @GET('^/v1/stream/list$')
    @GET('^/stream/list/v1$')
    def stream_list_v1(self, request):
        return api.streams_list(protocols=map(strng.to_text, filter(None, _request_arg(request, 'protocols', '').strip().lower().split(','))) or None)

    #------------------------------------------------------------------------------

    @GET('^/qu/l$')
    @GET('^/v1/queue/list$')
    @GET('^/queue/list/v1$')
    def queue_list_v1(self, request):
        return api.queues_list()

    @GET('^/qu/c/l$')
    @GET('^/v1/queue/consumer/list$')
    @GET('^/queue/consumer/list/v1$')
    def queue_consumer_list_v1(self, request):
        return api.queue_consumers_list()

    @GET('^/qu/p/l$')
    @GET('^/v1/queue/producer/list$')
    @GET('^/queue/producer/list/v1$')
    def queue_producer_list_v1(self, request):
        return api.queue_producers_list()

    @GET('^/qu/k/l$')
    @GET('^/v1/queue/keeper/list$')
    @GET('^/queue/keeper/list/v1$')
    def queue_keeper_list_v1(self, request):
        return api.queue_keepers_list()

    @GET('^/qu/ped/l$')
    @GET('^/v1/queue/peddler/list$')
    @GET('^/queue/peddler/list/v1$')
    def queue_peddler_list_v1(self, request):
        return api.queue_peddlers_list()

    @GET('^/qu/s/l$')
    @GET('^/v1/queue/stream/list$')
    @GET('^/queue/stream/list/v1$')
    def queue_stream_list_v1(self, request):
        return api.queue_streams_list()

    #------------------------------------------------------------------------------

    @GET('^/ev/l$')
    @GET('^/v1/event/list$')
    @GET('^/event/list/v1$')
    def event_list_v1(self, request):
        return api.events_list()

    @POST('^/ev/s/(?P<event_id>[^/]+)/$')
    @POST('^/v1/event/send/(?P<event_id>[^/]+)$')
    @POST('^/event/send/(?P<event_id>[^/]+)/v1$')
    def event_send_v1(self, request, event_id):
        return api.event_send(event_id, data=_request_data(request))

    @GET('^/ev/l/(?P<consumer_callback_id>[^/]+)/$')
    @GET('^/v1/event/listen/(?P<consumer_callback_id>[^/]+)$')
    @GET('^/event/listen/(?P<consumer_callback_id>[^/]+)/v1$')
    def event_listen_v1(self, request, consumer_callback_id):
        return api.event_listen(consumer_callback_id=consumer_callback_id)

    #------------------------------------------------------------------------------

    @GET('^/d/n/f$')
    @GET('^/v1/dht/node/find$')
    @GET('^/dht/node/find/v1$')
    def dht_node_find_v1(self, request):
        return api.dht_node_find(
            node_id_64=_request_arg(request, 'node_id_64', mandatory=False, default=None) or _request_arg(request, 'dht_id', mandatory=False, default=None),
            layer_id=int(_request_arg(request, 'layer_id', mandatory=False, default=0)),
        )

    @GET('^/d/u/r$')
    @GET('^/v1/dht/user/random$')
    @GET('^/dht/user/random/v1$')
    def dht_user_random_v1(self, request):
        return api.dht_user_random(
            layer_id=int(_request_arg(request, 'layer_id', mandatory=False, default=0)),
            count=int(_request_arg(request, 'count', mandatory=False, default=1)),
        )

    @GET('^/d/v/g$')
    @GET('^/v1/dht/value/get$')
    @GET('^/dht/value/get/v1$')
    def dht_value_get_v1(self, request):
        return api.dht_value_get(
            key=_request_arg(request, 'key', mandatory=True),
            record_type=_request_arg(request, 'record_type', mandatory=False, default='skip_validation'),
            layer_id=int(_request_arg(request, 'layer_id', mandatory=False, default=0)),
        )

    @POST('^/d/v/s$')
    @POST('^/v1/dht/value/set$')
    @POST('^/dht/value/set/v1$')
    def dht_value_set_v1(self, request):
        data = _request_data(request, mandatory_keys=['key', 'value'])
        return api.dht_value_set(
            key=data['key'],
            value=data['value'],
            expire=data.get('expire', None),
            record_type=data.get('record_type', 'skip_validation'),
            layer_id=int(data.get('layer_id', 0)),
        )

    @GET('^/d/d/d$')
    @GET('^/v1/dht/db/dump$')
    @GET('^/dht/db/dump/v1$')
    def dht_db_dump_v1(self, request):
        return api.dht_local_db_dump()

    #------------------------------------------------------------------------------

    @GET('^/b/i$')
    @GET('^/v1/blockchain/info$')
    @GET('^/blockchain/info/v1$')
    def blockchain_info(self, request):
        return api.blockchain_info()

    @GET('^/b/w/b$')
    @GET('^/v1/blockchain/wallet/balance$')
    @GET('^/blockchain/wallet/balance/v1$')
    def blockchain_wallet_balance(self, request):
        return api.blockchain_wallet_balance()

    @POST('^/b/t/s$')
    @POST('^/v1/blockchain/transaction/send$')
    @POST('^/blockchain/transaction/send/v1$')
    def blockchain_transaction_send(self, request):
        data = _request_data(request, mandatory_keys=['recipient', 'amount'])
        return api.blockchain_transaction_send(
            recipient=data['recipient'],
            amount=data['amount'],
            operation=data.get('operation', '') or '',
            data=data.get('data', '') or '',
        )

    @GET('^/b/b/p$')
    @GET('^/v1/blockchain/block/produce$')
    @GET('^/blockchain/block/produce/v1$')
    def blockchain_block_produce(self, request):
        return api.blockchain_block_produce()

    #------------------------------------------------------------------------------

    @GET('^/st/l$')
    @GET('^/v1/state/list$')
    @GET('^/v1/automat/list$')
    @GET('^/state/list/v1$')
    @GET('^/automat/list/v1$')
    def automat_list_v1(self, request):
        return api.automats_list()

    @GET('^/st/i$')
    @GET('^/v1/state/info$')
    @GET('^/v1/automat/info$')
    @GET('^/state/info/v1$')
    @GET('^/automat/info/v1$')
    def automat_info_v1(self, request):
        return api.automat_info(
            index=_request_arg(request, 'index', default=None, mandatory=False),
            automat_id=_request_arg(request, 'automat_id', default=None, mandatory=False),
        )

    @POST('^/st/e/start$')
    @POST('^/v1/state/events/start$')
    @POST('^/v1/automat/events/start$')
    @POST('^/state/events/start/v1$')
    @POST('^/automat/events/start/v1$')
    def automat_events_start_v1(self, request):
        data = _request_data(request)
        return api.automat_events_start(
            index=data.get('index', None),
            automat_id=data.get('automat_id', None),
            state_unchanged=bool(data.get('state_unchanged', '0') in ['1', 'true']),
        )

    @POST('^/st/e/stop$')
    @POST('^/v1/state/events/stop$')
    @POST('^/v1/automat/events/stop$')
    @POST('^/state/events/stop/v1$')
    @POST('^/automat/events/stop/v1$')
    def automat_events_stop_v1(self, request):
        data = _request_data(request)
        return api.automat_events_stop(
            index=data.get('index', None),
            automat_id=data.get('automat_id', None),
        )

    #------------------------------------------------------------------------------

    @ALL('^/*')
    def zzz_not_found(self, request):
        """
        This method is intended to return an error message when requested method was not found.
        Started with "zzz" because stuff is sorted alphabetically - so just to be able to put the regex on last place.
        """
        return api.ERROR('method %s:%s was not found' % (request.method, request.path))

    #------------------------------------------------------------------------------



#!/usr/bin/python
# deploy.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (deploy.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: deploy

Lets keep all software related data files in one place.
AppDataDir is a location of ".bitdust" folder basically.
However you can setup your donated location in another place: USB-stick, second hard disk, etc ...

Linux: /home/$USER/.bitdust
WindowsXP: c:\\Document and Settings\\[user]\\.bitdust
Windows7 and later: c:\\Users\\[user]\\.bitdust
MacOS: /Users/$USER/.bitdust

"""

#------------------------------------------------------------------------------

import os
import sys
import platform

#------------------------------------------------------------------------------

_AppDataDirPath = None
_CurrentNetwork = None

#------------------------------------------------------------------------------


def print_text(msg, nl='\n'):
    """
    Send some output to the console.
    """
    sys.stdout.write(msg + nl)
    sys.stdout.flush()


def get_executable_location():
    """
    Returns path to the folder from where current process was executed.
    """
    try:
        source_dir = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
    except:
        source_dir = os.path.dirname(os.path.dirname(os.path.abspath(sys.argv[0])))
    return source_dir


def appdata_location_file_path():
    """
    You can configure BitDust software to use another place for data folder.

    Say you want to store BitDust files on another disk. In the binary
    folder file "appdata" can be created and it will keep the path to
    the data folder.
    """
    return os.path.join(get_executable_location(), 'appdata')


def current_base_dir():
    """
    Returns currently known base location.
    """
    global _AppDataDirPath
    return _AppDataDirPath


def set_base_dir(new_path):
    """
    Rewrite currently known location of base dir with new path.
    """
    global _AppDataDirPath
    _AppDataDirPath = new_path


def current_network():
    global _CurrentNetwork
    return _CurrentNetwork


def default_base_dir_portable():
    """
    A portable method to get the default data folder location.
    """
    if platform.uname()[0] == 'Windows':
        return os.path.join(os.path.expanduser('~'), '.bitdust')

    elif platform.uname()[0] == 'Linux':
        if 'ANDROID_ARGUMENT' in os.environ or 'ANDROID_ROOT' in os.environ:
            from android.storage import app_storage_path  # @UnresolvedImport
            return os.path.join(app_storage_path(), '.bitdust')

        return os.path.join(os.path.expanduser('~'), '.bitdust')

    elif platform.uname()[0] == 'Darwin':
        return os.path.join(os.path.expanduser('~'), '.bitdust')

    elif 'ANDROID_ARGUMENT' in os.environ or 'ANDROID_ROOT' in os.environ:
        from android.storage import app_storage_path  # @UnresolvedImport
        return os.path.join(app_storage_path(), '.bitdust')

    return os.path.join(os.path.expanduser('~'), '.bitdust')


def init_current_network(name=None, base_dir=None):
    global _CurrentNetwork
    base_dir = base_dir or current_base_dir()
    if name:
        if name == 'current_network':
            raise Exception('invalid network name')
        _CurrentNetwork = name
        open(os.path.join(base_dir, 'current_network'), 'w').write(_CurrentNetwork)
        return
    if _CurrentNetwork:
        return
    try:
        cur_network = open(os.path.join(base_dir, 'current_network'), 'r').read().strip()
    except:
        cur_network = 'default'
    if not os.path.isdir(os.path.join(base_dir, cur_network)):
        cur_network = 'default'
    _CurrentNetwork = cur_network
    open(os.path.join(base_dir, 'current_network'), 'w').write(_CurrentNetwork)


def init_base_dir(base_dir=None, network_name=None):
    """
    Do some validation and create needed data folders if they are not exist
    yet.

    You can specify another location for data files.
    """
    global _AppDataDirPath
    global _CurrentNetwork

    # if we already know the place - we are done
    if base_dir:
        _AppDataDirPath = base_dir
        if not os.path.exists(_AppDataDirPath):
            os.makedirs(_AppDataDirPath, 0o777)
        init_current_network(name=network_name)
        return _AppDataDirPath

    # if location was already known - no need to check again
    if _AppDataDirPath is not None:
        init_current_network(name=network_name)
        return _AppDataDirPath

    # if we have a file "appdata" in current folder - read location path from there
    appdata_path = appdata_location_file_path()
    if os.path.isfile(appdata_path):
        if os.path.isfile(appdata_path) and os.access(appdata_path, os.R_OK):
            infile = open(appdata_path, 'r')
            path = infile.read().strip()
            infile.close()
            if path:
                path = os.path.abspath(path)
                if not os.path.isdir(path):
                    os.makedirs(path, 0o777)
                _AppDataDirPath = path
                init_current_network(name=network_name)
                return _AppDataDirPath

    # get the default place for that machine
    default_path = default_base_dir_portable()

    # we can use folder ".bitdust" placed on the same level with binary folder:
    # /..
    #   /.bitdust - data files
    #   /bitdust  - source files
    path1 = str(os.path.abspath(os.path.join(get_executable_location(), '..', '.bitdust')))
    # and default path will have lower priority
    path2 = default_path

    # if default path exists - use it
    if os.path.isdir(path2):
        _AppDataDirPath = path2
    # but ".bitdust" folder on same level will have higher priority
    if os.path.isdir(path1):
        _AppDataDirPath = path1
    # use default path if nothing existing yet
    if not _AppDataDirPath:
        _AppDataDirPath = default_path

    # if we did not found "current_network" file - use default path, new copy of BitDust
    if not os.path.isdir(os.path.join(current_base_dir(), 'current_network')):
        _AppDataDirPath = path2
        if not os.path.exists(_AppDataDirPath):
            os.makedirs(_AppDataDirPath)
        init_current_network(name=network_name)
        return _AppDataDirPath

    # if we did not found our key - use default path, new copy of BitDust
    if not os.access(os.path.join(current_base_dir(), current_network(), 'metadata', 'mykeyfile'), os.R_OK) or \
        not os.access(os.path.join(current_base_dir(), current_network(), 'metadata', 'mykeyfile_location'), os.R_OK):
        _AppDataDirPath = path2
        if not os.path.exists(_AppDataDirPath):
            os.makedirs(_AppDataDirPath, 0o777)
        init_current_network(name=network_name)
        return _AppDataDirPath

    # if we did not found our identity - use default path, new copy of BitDust
    if not os.access(os.path.join(current_base_dir(), current_network(), 'metadata', 'localidentity'), os.R_OK):
        _AppDataDirPath = path2
        if not os.path.exists(_AppDataDirPath):
            os.makedirs(_AppDataDirPath)
        init_current_network(name=network_name)
        return _AppDataDirPath

    init_current_network(name=network_name)
    # seems we found needed files in a path1 - let us use this as a base dir
    return _AppDataDirPath


#------------------------------------------------------------------------------


def run(args):
    """
    Creates virtual environment
    """
    status = 1
    on_windows = platform.uname()[0] == 'Windows'
    on_mac = platform.uname()[0] == 'Darwin'
    on_linux = platform.uname()[0] == 'Linux'
    source_dir = get_executable_location()
    init_base_dir()
    base_dir = current_base_dir()
    if on_windows and os.path.isfile(os.path.join(base_dir, 'shortpath.txt')):
        base_dir = open(os.path.join(base_dir, 'shortpath.txt')).read().strip()
    venv_path = os.path.join(base_dir, 'venv')
    if len(args) > 1 and not os.path.exists(args[1]) and os.path.isdir(os.path.dirname(args[1])):
        venv_path = args[1]
    pip_bin = '{}/bin/pip'.format(venv_path)
    script_path = os.path.join(base_dir, 'bitdust')

    if os.path.exists(venv_path):
        print_text('\n***** Clean up existing Python virtual environment in "%s"' % venv_path)
        if on_windows:
            status = os.system('rmdir /S /Q "{}"'.format(venv_path))
        else:
            status = os.system('rm -rf {}'.format(venv_path))
        if status != 0:
            print_text('\n***** Clean up of existing virtual environment files failed!\n')
            return status

    current_python = sys.executable
    print_text('\n***** Current Python executable is {}'.format(current_python))

    print_text('\n***** Creating virtual environment in "%s"' % venv_path)
    make_venv_cmd = 'virtualenv -p {} {}'.format(current_python, venv_path)
    if on_windows:
        python_exe = '"%s"' % os.path.join(base_dir, 'python', 'python.exe')
        if not os.path.exists(python_exe):
            python_exe = current_python
        make_venv_cmd = '{} -m virtualenv --system-site-packages {}'.format(python_exe, venv_path)
    if on_mac:
        if sys.version_info >= (3, 11):
            make_venv_cmd = '{} -m venv --clear --copies {}'.format(current_python, venv_path)
        else:
            make_venv_cmd = '{} -m virtualenv --clear --always-copy {}'.format(current_python, venv_path)

    print_text('\n***** Executing "{}"'.format(make_venv_cmd))
    status = os.system(make_venv_cmd)
    if on_mac and status != 0:
        make_venv_cmd = 'virtualenv -p {} {}'.format(current_python, venv_path)
        status = os.system(make_venv_cmd)
    if on_linux and status != 0:
        make_venv_cmd = '{} -m pip install -q virtualenv'.format(current_python)
        status = os.system(make_venv_cmd)
        make_venv_cmd = '{} -m virtualenv --clear --always-copy {}'.format(current_python, venv_path)
        status = os.system(make_venv_cmd)

    if status != 0:
        print_text('\n***** Failed to create virtual environment, please check/install virtualenv package\n')
        print_text('\n***** Please try to install virtualenv package manually:\n')
        print_text('    {} -m pip install virtualenv\n\n'.format(current_python))
        return status

    if on_windows or on_mac:
        pass
    else:
        print_text('\n***** Installing/Upgrading pip in "%s"' % venv_path)
        status = os.system('{} install -U pip'.format(pip_bin))
        if status != 0:
            # print_text('\n***** Failed to install latest pip version, please check/install latest pip version manually\n')
            # return status
            pass

    if on_mac:
        print_text('\n***** Updating setuptools version in "%s"' % venv_path)
        status = os.system('{} install setuptools'.format(pip_bin))
        if status != 0:
            print_text('\n***** Failed to install/upgrade setuptools, please check/install setuptools manually\n')
            return status

    requirements_txt = os.path.join(source_dir, 'requirements.txt')
    print_text('\n***** Installing BitDust requirements from "%s"' % (requirements_txt))
    requirements_cmd = '{} install -q -r "{}"'.format(pip_bin, requirements_txt)
    if on_windows:
        venv_python_path = os.path.join(base_dir, 'venv', 'Scripts', 'python.exe')
        requirements_cmd = '{} -m pip install -q -r "{}"'.format(venv_python_path, requirements_txt)
    if on_mac:
        venv_python_path = os.path.join(base_dir, 'venv', 'bin', 'python')
        requirements_cmd = '{} -m pip install -q -r "{}"'.format(venv_python_path, requirements_txt)

    print_text('\n***** Executing "{}"'.format(requirements_cmd))
    status = os.system(requirements_cmd)

    if on_windows and status == 0:
        extra_requirements_cmd = '{} -m pip install -q pywin32 pypiwin32'.format(venv_python_path)
        status = os.system(extra_requirements_cmd)

    if status != 0:
        # TODO: try to detect package manager on target OS and give more correct info: debian/mandrake/OSX
        # TODO: on Windows it could be that some of the .wheels are not available - need to provide a different message
        depends = [
            'gcc',
            'build-essential',
            'libssl-dev',
            'libffi-dev',
            'python3-dev',
            'python3-virtualenv',
        ]
        print_text('\n***** Please try to install those binary packages manually:\n')
        print_text('    %s\n\n' % (' '.join(depends)))
        return status

    # TODO: on Windows this script should create .bat file instead
    script = u'#!/bin/sh\n'
    script += u'# This is a short shell script to create an alias in OS for BitDust software.\n'
    script += u'# NOTICE: BitDust software do not require root permissions to run, please start it as normal user.\n\n'
    script += u'{}/bin/python {}/bitdust.py "$@"\n\n'.format(venv_path, source_dir)
    fil = open(script_path, mode='w')
    fil.write(script)
    fil.close()
    os.chmod(script_path, 0o775)

    print_text('\n***** BitDust environment files created successfully in {}\n'.format(base_dir))
    print_text('To create system-wide shell command you can add folder {} to your PATH, or create a symbolic link:\n'.format(script_path))
    print_text('    sudo ln -s -f {} /usr/local/bin/bitdust\n\n'.format(script_path))
    print_text('To run BitDust without system-wide alias use that executable file:\n\n    {}\n\n'.format(script_path))
    print_text('Learn more about how to use the software via command line:\n\n    bitdust help\n\n')
    print_text('Welcome to BitDust!\n\n')
    return 0



"""

test cmd for the hn_last_block_ts command.


Usage:

python3 cmd_hn_last_block_ts

"""

import connections
import json
import socks

__version__ = '0.0.1'

ORIGIN_OF_TIME = 1534716000  # Real Origin: August 20
POS_SLOT_TIME_MIN = 3  # Real world setting?
POS_SLOT_TIME_SEC = POS_SLOT_TIME_MIN*60
MAX_ROUND_SLOTS = 19  # Real world. 19+1 = 20 , 3x20 = 60 (round time)
END_ROUND_SLOTS = 1
# Round time in seconds
ROUND_TIME_SEC = POS_SLOT_TIME_SEC*(MAX_ROUND_SLOTS + END_ROUND_SLOTS)


def hn_last_block_ts():
    s = socks.socksocket()
    s.settimeout(10)
    s.connect(('127.0.0.1', 5658))
    # Last param is ip, to get feed of a specific ip, False for all.
    connections.send(s, 'HN_last_block_ts')
    res = connections.receive(s)
    return res


if __name__ == '__main__':
    res_as_dict = hn_last_block_ts()
    print('Answer (<=0 means fail):')
    print(json.dumps(res_as_dict, indent=2))
